import astropy.units as u
import numpy as np
import scipy.special
import named_arrays as na

__all__ = [
    "charge_diffusion",
    "mean_charge_capture",
]


def charge_diffusion(
    absorption: u.Quantity | na.AbstractScalar,
    thickness_substrate: u.Quantity | na.AbstractScalar,
    thickness_depletion: u.Quantity | na.AbstractScalar,
) -> na.AbstractScalar:
    r"""
    The standard deviation of the charge diffusion in a backilluminated CCD
    given by :cite:t:`Janesick2001`.

    Parameters
    ----------
    absorption
        The absorption coefficient of the light-sensitive layer for the
        incident photon.
    thickness_substrate
        The thickness of the light-sensitive region of the imaging sensor.
    thickness_depletion
        The thickness of the depletion region of the imaging sensor.

    Examples
    --------

    Plot the width of the charge diffusion kernel as a function of wavelength
    and energy for the sensor parameters in :cite:t:`Heymes2020`.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Define a grid of wavelengths
        wavelength = na.geomspace(1, 10000, axis="wavelength", num=1001) * u.AA

        # Convert the grid to energies as well
        energy = wavelength.to(u.eV, equivalencies=u.spectral())

        # Load the optical properties of silicon
        si = optika.chemicals.Chemical("Si")

        # Retrieve the absorption coefficient of silicon
        # for the given wavelenghts.
        absorption = si.absorption(wavelength)

        # Compute the charge diffusion
        width_diffusion = optika.sensors.charge_diffusion(
            absorption=absorption,
            thickness_substrate=14 * u.um,
            thickness_depletion=2.4 * u.um,
        )

        # Plot the charge diffusion as a function
        # of wavelength and energy
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            ax2 = ax.twiny()
            ax2.invert_xaxis()
            na.plt.plot(
                wavelength,
                width_diffusion,
                ax=ax,
            )
            na.plt.plot(
                energy,
                width_diffusion,
                ax=ax2,
                linestyle="None",
            )
            ax.set_xscale("log")
            ax2.set_xscale("log")
            ax.set_xlabel(f"wavelength ({ax.get_xlabel()})")
            ax.set_ylabel(f"charge diffusion ({ax.get_ylabel()})")

    Notes
    -----

    The standard deviation of the charge diffusion kernel is given by
    :cite:t:`Janesick2001` as

    .. math::

        \sigma_\text{cd}(x) = \begin{cases}
            x_{ff} \sqrt{1 - \frac{x}{x_{ff}}}, & 0 < x < x_{ff} \\
            0, & x_{ff} < x < x_s
        \end{cases}

    where :math:`x` is the distance from the back surface at which the photon
    is absorbed,

    .. math::

        x_{ff} = x_s - x_d

    is the thickness of the field-free region of the sensor,
    :math:`x_s` is the total thickness of the light-sensitive region,
    and :math:`x_d` is the thickness of the depletion region.

    The `average` variance of the charge diffusion kernel is then
    the weighted average,

    .. math::

        \overline{\sigma}_\text{cd}^2 &= \dfrac{\displaystyle \int_0^{x_s} \left( \sigma_\text{cd}(x) \right)^2 e^{-\alpha x} dx}
                                               {\displaystyle \int_0^{x_s} e^{-\alpha x} dx} \\[1mm]
                                      &= \dfrac{\displaystyle \int_0^{x_{ff}} x_{ff}^2 \left( 1 - \frac{x}{x_{ff}} \right) e^{-\alpha x} dx}
                                               {\displaystyle \int_0^{x_s} e^{-\alpha x} dx} \\[1mm]
                                      &= \dfrac{x_{ff} \left( \alpha x_{ff} + e^{-\alpha x_{ff}} - 1 \right)}
                                               {\alpha \left( 1 - e^{-\alpha x_s} \right)}

    where :math:`\alpha` is the absorption coefficient of the light-sensitive layer.
    """
    s = thickness_substrate

    f = s - thickness_depletion

    a = absorption

    numerator = f * (a * f + np.exp(-a * f) - 1)
    denominator = a * (1 - np.exp(-a * s))

    result = np.sqrt(numerator / denominator).to(u.um)

    return result


def mean_charge_capture(
    width_diffusion: u.Quantity | na.AbstractScalar,
    width_pixel: u.Quantity | na.AbstractScalar,
) -> na.AbstractScalar:
    r"""
    A function to compute the mean charge capture :cite:p:`Stern2004`,
    the fraction of charge from each photon event retained in the central pixel.

    Parameters
    ----------
    width_diffusion
        The standard deviation of the charge diffusion kernel.
    width_pixel
        The width of a pixel on the sensor.

    Examples
    --------

    Plot the mean charge capture as a function of wavelength
    and energy for the sensor parameters in :cite:t:`Heymes2020`.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Define a grid of wavelengths
        wavelength = na.geomspace(1, 10000, axis="wavelength", num=1001) * u.AA

        # Convert the grid to energies as well
        energy = wavelength.to(u.eV, equivalencies=u.spectral())

        # Load the optical properties of silicon
        si = optika.chemicals.Chemical("Si")

        # Retrieve the absorption coefficient of silicon
        # for the given wavelenghts.
        absorption = si.absorption(wavelength)

        # Compute the charge diffusion
        width_diffusion = optika.sensors.charge_diffusion(
            absorption=absorption,
            thickness_substrate=14 * u.um,
            thickness_depletion=2.4 * u.um,
        )

        # Compute the mean charge capture
        mcc = optika.sensors.mean_charge_capture(
            width_diffusion=width_diffusion,
            width_pixel=16 * u.um,
        )

        # Plot the mean charge capture as a function
        # of wavelength and energy
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            ax2 = ax.twiny()
            ax2.invert_xaxis()
            na.plt.plot(
                wavelength,
                mcc,
                ax=ax,
            )
            na.plt.plot(
                energy,
                mcc,
                ax=ax2,
                linestyle="None",
            )
            ax.set_xscale("log")
            ax2.set_xscale("log")
            ax.set_xlabel(f"wavelength ({ax.get_xlabel()})")
            ax2.set_xlabel(f"energy ({ax2.get_xlabel()})")
            ax.set_ylabel(f"mean charge capture")

    Notes
    -----
    Naively, the mean charge capture (MCC) is the integral of the charge
    diffusion kernel over the extent of a pixel.
    However, since a photon can strike anywhere within the central pixel,
    the charge diffusion kernel should be convolved with a rectangle function
    the width of a pixel before integrating.
    So, our definition for the MCC is

    .. math::

        P_\text{MCC} = \left\{ \frac{1}{d} \int_{-d/2}^{d/2} \left[ K(x') * \Pi \left( \frac{x'}{d} \right) \right](x) \, dx \right\}^2,

    where :math:`K(x)` is the charge diffusion kernel,
    :math:`\Pi(x)` is the `rectangle function <https://en.wikipedia.org/wiki/Rectangular_function>`_,
    and :math:`d` is the width of a pixel.
    If we assume that the charge diffusion kernel is a Gaussian with standard
    deviation :math:`\sigma`,

    .. math::

        K(x) = \frac{1}{\sqrt{2\pi} \sigma} \exp \left( -\frac{x^2}{2 \sigma^2} \right),

    then we can analytically solve for the MCC,

    .. math::

        P_\text{MCC} &= \left\{ \frac{1}{2d} \int_{-d/2}^{d/2} \left[ \text{erf} \left( \frac{d - 2x}{2 \sqrt{2} \sigma} \right) + \text{erf} \left( \frac{d + 2x}{2 \sqrt{2} \sigma} \right) \right] dx \right\}^2 \\
                     &= \left\{ \sqrt{\frac{2}{\pi}} \frac{\sigma}{d} \left[ \exp \left( -\frac{d^2}{2 \sigma^2} \right) - 1 \right] + \text{erf} \left( \frac{d}{\sqrt{2} \sigma} \right) \right\}^2,

    where :math:`\text{erf}(x)` is the `error function <https://en.wikipedia.org/wiki/Error_function>`_.
    """
    a = width_pixel / width_diffusion

    t1 = np.sqrt(2 / np.pi) * (np.exp(-np.square(a) / 2) - 1) / a
    t2 = scipy.special.erf(a / np.sqrt(2))

    result = np.square(t1 + t2)

    return result.to(u.dimensionless_unscaled)



from __future__ import annotations
import dataclasses
import named_arrays as na
import optika
from . import RayVectorArray

__all__ = [
    "AbstractRayFunctionArray",
    "RayFunctionArray",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractRayFunctionArray(
    na.AbstractFunctionArray,
):
    @property
    def type_explicit(self) -> type[RayFunctionArray]:
        return RayFunctionArray


@dataclasses.dataclass(eq=False, repr=False)
class RayFunctionArray(
    AbstractRayFunctionArray,
    na.FunctionArray[
        optika.vectors.ObjectVectorArray,
        RayVectorArray,
    ],
):
    pass



import abc
import functools
import dataclasses
import numpy as np
import scipy.optimize
import astropy.units as u
import astropy.constants
import named_arrays as na
import optika
from ._depletion import AbstractDepletionModel

__all__ = [
    "energy_bandgap",
    "energy_electron_hole",
    "quantum_yield_ideal",
    "absorbance",
    "charge_collection_efficiency",
    "quantum_efficiency_effective",
    "probability_measurement",
    "electrons_measured",
    "signal",
    "AbstractImagingSensorMaterial",
    "IdealImagingSensorMaterial",
    "AbstractCCDMaterial",
    "AbstractBackilluminatedCCDMaterial",
    "AbstractStern1994BackilluminatedCCDMaterial",
]

energy_bandgap = 1.12 * u.eV
"""the bandgap energy of silicon"""

energy_electron_hole = 3.65 * u.eV
"""
the high-energy limit of the energy required to create an electron-hole pair
in silicon at room temperature
"""

_thickness_oxide = 50 * u.AA
_thickness_substrate = 7 * u.um
_thickness_implant = 2317 * u.AA
_cce_backsurface = 0.21
_fano_noise = 0.1 * u.electron / u.photon


def quantum_yield_ideal(
    wavelength: u.Quantity | na.AbstractScalar,
) -> na.AbstractScalar:
    r"""
    Calculate the ideal quantum yield of a silicon detector for a given
    wavelength.

    Parameters
    ----------
    wavelength
        the wavelength of the incident photons

    Notes
    -----
    The quantum yield is the number of electron-hole pairs produced per photon.
    This is also known in the literature as the pair-production energy.

    The ideal quantum yield is given in :cite:t:`Janesick2001` as:

    .. math::

        \text{QY}(\epsilon) = \begin{cases}
            0, & 0 \leq \epsilon < E_\text{g}\\
            1, &  E_\text{g} \leq \epsilon < E_\text{e-h} \\
            \epsilon / E_\text{e-h}, & E_\text{e-h} \leq \epsilon < \infty,
        \end{cases},

    where :math:`\epsilon` is the energy of the incident photon,
    :math:`E_\text{g} = 1.12\;\text{eV}` is the bandgap energy of silicon,
    and :math:`E_\text{e-h} = 3.65\;\text{eV}` is the energy required to
    generate 1 electron-hole pair in silicon at room temperature.

    Examples
    --------

    Plot the quantum yield vs wavelength

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of wavelengths
        wavelength = na.geomspace(100, 100000, axis="wavelength", num=101) << u.AA

        # Compute the quantum yield
        qy = optika.sensors.quantum_yield_ideal(wavelength)

        # Plot the quantum yield vs wavelength
        fig, ax = plt.subplots()
        na.plt.plot(wavelength, qy, ax=ax);
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel(f"quantum yield ({qy.unit:latex_inline})");
    """
    energy = wavelength.to(u.eV, equivalencies=u.spectral())

    result = energy / energy_electron_hole
    result = np.where(energy > energy_electron_hole, result, 1)
    result = np.where(energy > energy_bandgap, result, 0)

    return result * u.electron / u.photon


def absorbance(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar = 1,
    n: float | na.AbstractScalar = 1,
    thickness_oxide: u.Quantity | na.AbstractScalar = _thickness_oxide,
    thickness_substrate: u.Quantity | na.AbstractScalar = _thickness_substrate,
    chemical_oxide: str | optika.chemicals.AbstractChemical = "SiO2",
    chemical_substrate: str | optika.chemicals.AbstractChemical = "Si",
    roughness_oxide: u.Quantity | na.AbstractScalar = 0 * u.nm,
    roughness_substrate: u.Quantity | na.AbstractScalar = 0 * u.nm,
) -> optika.vectors.PolarizationVectorArray:
    """
    The fraction of incident energy absorbed by the light-sensitive
    region of the sensor

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The component of the incident light's propagation direction antiparallel
        to the surface normal of the sensor.
        Default is normal incidence.
    n
        The index of refraction in the ambient medium.
    thickness_oxide
        The thickness of the oxide layer on the illuminated surface of the sensor.
        Default is the value given in :cite:t:`Stern1994`.
    thickness_substrate
        The thickness of the light-sensitive substrate layer.
        Default is the value given in :cite:t:`Stern1994`.
    chemical_oxide
        The chemical formula of the oxide layer on the illuminated surface of the sensor.
        Default is silicon dioxide.
    chemical_substrate
        The chemical formula of the light-sensitive portion of the sensor.
        Default is silicon.
    roughness_oxide
        The RMS roughness the oxide layer surface.
    roughness_substrate
        The RMS roughness of the substrate surface.

    Examples
    --------

    Plot the absorbance as a function of wavelength.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define a grid of wavelengths
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Compute the absorbance vs wavelength
        absorbance = optika.sensors.absorbance(
            wavelength=wavelength,
        )

        # Plot the effective and maximum quantum efficiency
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.plot(
            wavelength,
            absorbance.average,
            ax=ax,
        );
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("incident energy fraction");
    """
    if not isinstance(chemical_oxide, optika.chemicals.AbstractChemical):
        chemical_oxide = optika.chemicals.Chemical(chemical_oxide)

    if not isinstance(chemical_substrate, optika.chemicals.AbstractChemical):
        chemical_substrate = optika.chemicals.Chemical(chemical_substrate)

    result = optika.materials.layer_absorbance(
        index=1,
        wavelength=wavelength,
        direction=direction,
        n=n,
        layers=[
            optika.materials.Layer(
                chemical=chemical_oxide,
                thickness=thickness_oxide,
                interface=optika.materials.profiles.ErfInterfaceProfile(
                    width=roughness_oxide,
                ),
            ),
            optika.materials.Layer(
                chemical=chemical_substrate,
                thickness=thickness_substrate,
                interface=optika.materials.profiles.ErfInterfaceProfile(
                    width=roughness_substrate,
                ),
            ),
        ],
    )

    return np.real(result)


def charge_collection_efficiency(
    absorption: u.Quantity | na.AbstractScalar,
    thickness_implant: u.Quantity | na.AbstractScalar = _thickness_implant,
    cce_backsurface: u.Quantity | na.AbstractScalar = _cce_backsurface,
    cos_incidence: float | na.AbstractScalar = 1,
) -> na.AbstractScalar:
    r"""
    Compute the average charge collection efficiency using the differential
    charge collection efficiency profile described in :cite:t:`Stern1994`.

    Parameters
    ----------
    absorption
        The absorption coefficient of the light-sensitive material for the
        wavelength of interest.
    thickness_implant
        The thickness of the implant layer, the layer where recombination can
        occur.
        Default is the value given in :cite:t:`Stern1994`.
    cce_backsurface
        The differential charge collection efficiency on the back surface
        of the sensor.
        Default is the value given in :cite:t:`Stern1994`.
    cos_incidence
        The cosine of the angle of the incident light's propagation direction
        inside the substrate with the surface normal

    Examples
    --------

    Plot the charge collection efficiency as a function of wavelength.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define a grid of wavelengths
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Compute the absorption coefficient for silicon
        absorption = optika.chemicals.Chemical("Si").absorption(wavelength)

        # Compute the CCE vs wavelength
        cce = optika.sensors.charge_collection_efficiency(
            absorption=absorption,
        )

        # Plot the effective and maximum quantum efficiency
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.plot(
            wavelength,
            cce,
            ax=ax,
        );
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("charge collection efficiency");

    Notes
    -----

    The charge collection efficiency is the fraction of photoelectrons that
    are measured by the sensor :cite:p:`Janesick2001`,
    and is an important component of the quantum efficiency of the sensor

    In :cite:t:`Stern1994`, the authors define a differential charge collection
    efficiency, :math:`\eta(z)`, which is the probability that a photoelectron
    resulting from a photon absorbed at a depth :math:`z` will be measured by
    the sensor.
    In principle, :math:`\eta(z)` depends on the exact implant profile on the
    backsurface of the sensor, however :cite:t:`Stern1994` and :cite:t:`Boerner2012`
    have shown that a piecewise-linear approximation of :math:`\eta(z)`,

    .. math::
        :label: differential-cce

        \eta(z) = \begin{cases}
            \eta_0 + (1 - \eta_0) z / W, & 0 < z < W \\
            1, & W < z < D,
        \end{cases}

    is sufficient, given the uncertainties in the optical constants involved.

    The total charge collection efficiency is then the average value of
    :math:`\eta(z)` weighted by the probability of absorbing a photon at a
    depth :math:`z`,

    .. math::
        :label: cce-definition

        \text{CCE}(\lambda) = \frac{\int_0^\infty \eta(z) e^{-\alpha z} \, dz}
                               {\int_0^\infty e^{-\alpha z} \, dz}.

    Plugging Equation :eq:`differential-cce` into Equation :eq:`cce-definition`
    and integrating yields

    .. math::
        :label: cce

        \text{CCE}(\lambda) = \eta_0 + \left( \frac{1 - \eta_0}{\alpha W} \right)(1 - e^{-\alpha W}).

    Equation :eq:`cce` is equivalent to the term in curly braces of Equation 11 in :cite:t:`Stern1994`,
    with the addition of an :math:`e^{-\alpha W}` term which represents photons
    absorbed inside the epitaxial layer but outside the implant layer.

    Equation :eq:`cce` is only valid for normally-incident light.
    We can generalize it to obliquely-incident light by making the substitution

    .. math::
        :label: x-oblique

        x \rightarrow \frac{x}{\cos \theta}

    where :math:`\theta` is the angle between the propagation direction
    inside the silicon substrate and the normal vector.

    Substituting :eq:`x-oblique` into Equation :eq:`cce` and solving yields

    .. math::
        :label: eqe-oblique

        \text{CCE}(\lambda, \theta) =
            \eta_0
            + \left( \frac{1 - \eta_0}{\alpha W \sec \theta} \right) (1 - e^{-\alpha W \sec \theta})
    """
    z0 = absorption * thickness_implant / np.real(cos_incidence)
    exp_z0 = np.exp(-z0)

    term_1 = cce_backsurface
    term_2 = ((1 - cce_backsurface) / z0) * (1 - exp_z0)

    return term_1 + term_2


def quantum_efficiency_effective(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: None | na.AbstractCartesian3dVectorArray = None,
    n: float | na.AbstractScalar = 1,
    thickness_oxide: u.Quantity | na.AbstractScalar = _thickness_oxide,
    thickness_implant: u.Quantity | na.AbstractScalar = _thickness_implant,
    thickness_substrate: u.Quantity | na.AbstractScalar = _thickness_substrate,
    cce_backsurface: u.Quantity | na.AbstractScalar = _cce_backsurface,
    chemical_oxide: str | optika.chemicals.AbstractChemical = "SiO2",
    chemical_substrate: str | optika.chemicals.AbstractChemical = "Si",
    roughness_oxide: u.Quantity | na.AbstractScalar = 0 * u.nm,
    roughness_substrate: u.Quantity | na.AbstractScalar = 0 * u.nm,
    normal: None | na.AbstractCartesian3dVectorArray = None,
) -> na.AbstractScalar:
    r"""
    Calculate the effective quantum efficiency of a backilluminated detector.

    Parameters
    ----------
    wavelength
        The wavelength of the incident light in vacuum.
    direction
        The propagation direction of the incident light in the ambient medium.
        If :obj:`None` (default), normal incidence (:math:`\hat{z}`) is assumed.
    n
        The complex index of refraction of the ambient medium.
    thickness_oxide
        The thickness of the oxide layer on the back surface of the sensor.
        Default is the value given in :cite:t:`Stern1994`.
    thickness_implant
        The thickness of the implant layer.
        Default is the value given in :cite:t:`Stern1994`.
    thickness_substrate
        The thickness of the silicon substrate.
        Default is the value given in :cite:t:`Stern1994`.
    cce_backsurface
        The differential charge collection efficiency on the back surface
        of the sensor.
        Default is the value given in :cite:t:`Stern1994`.
    chemical_oxide
        The chemical composition of the oxide layer.
        The default is to assume the oxide layer is silicon dioxide.
    chemical_substrate
        Optional complex refractive index of the implant region and substrate.
        The default is to assume the substrate is made from silicon.
    roughness_oxide
        The RMS roughness the oxide layer surface.
    roughness_substrate
        The RMS roughness of the substrate surface.
    normal
        The vector perpendicular to the surface of the sensor.
        If :obj:`None`, then the normal is assumed to be :math:`-\hat{z}`

    Examples
    --------
    Reproduce Figure 12 from :cite:t:`Stern1994`, the modeled quantum efficiency
    of a Tektronix TK512CB :math:`512 \times 512` pixel backilluminated CCD.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import numpy as np
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of wavelengths with which to sample the EQE
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Compute the effective quantum efficiency
        eqe = optika.sensors.quantum_efficiency_effective(
            wavelength=wavelength,
        )

        # Compute the maximum theoretical quantum efficiency
        eqe_max = optika.sensors.quantum_efficiency_effective(
            wavelength=wavelength,
            cce_backsurface=1,
        )

        # Plot the effective and maximum quantum efficiency
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.plot(
            wavelength,
            eqe,
            ax=ax,
            label="effective quantum efficiency",
        );
        na.plt.plot(
            wavelength,
            eqe_max,
            ax=ax,
            label="maximum quantum efficiency",
        );
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("efficiency");
        ax.legend();

    Plot the EQE as a function of wavelength for normal and oblique incidence

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import numpy as np
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define an array of wavelengths with which to sample the EQE
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Define the incidence directions
        angle = na.linspace(0, 30, axis="angle", num=2) * u.deg
        direction = na.Cartesian3dVectorArray(
            x=np.sin(angle),
            y=0,
            z=np.cos(angle),
        )

        eqe = optika.sensors.quantum_efficiency_effective(
            wavelength=wavelength,
            direction=direction,
        )

        # Plot the results
        fig, ax = plt.subplots(constrained_layout=True)
        angle_str = angle.value.astype(str).astype(object)
        na.plt.plot(
            wavelength,
            eqe,
            ax=ax,
            axis="wavelength",
            label=r"$\theta$ = " + angle_str + f"{angle.unit:latex_inline}",
        );
        ax.set_xscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel("efficiency");
        ax.legend();

    Notes
    -----
    :cite:t:`Stern1994` defines the effective quantum efficiency as

    .. math::
        :label: eqe

        \text{EQE}(\lambda) = A(\lambda) \times \text{CCE}(\lambda),

    where :math:`A(\lambda)` is the absorbtivity of the epitaxial silicon layer
    for a given wavelength :math:`\lambda`,
    and :math:`\text{CCE}(\lambda)` is the charge collection efficiency
    (computed by :func:`charge_collection_efficiency`).
    """
    if direction is None:
        direction = na.Cartesian3dVectorArray(0, 0, 1)

    if not isinstance(chemical_oxide, optika.chemicals.AbstractChemical):
        chemical_oxide = optika.chemicals.Chemical(chemical_oxide)

    if not isinstance(chemical_substrate, optika.chemicals.AbstractChemical):
        chemical_substrate = optika.chemicals.Chemical(chemical_substrate)

    if normal is None:
        normal = na.Cartesian3dVectorArray(0, 0, -1)

    direction = -direction @ normal

    absorbance_substrate = absorbance(
        wavelength=wavelength,
        direction=direction,
        n=n,
        thickness_oxide=thickness_oxide,
        thickness_substrate=thickness_substrate,
        chemical_oxide=chemical_oxide,
        chemical_substrate=chemical_substrate,
        roughness_oxide=roughness_oxide,
        roughness_substrate=roughness_substrate,
    )

    n_substrate = chemical_substrate.n(wavelength)
    wavenumber_substrate = np.imag(n_substrate)
    absorption_substrate = 4 * np.pi * wavenumber_substrate / wavelength

    direction_substrate = optika.materials.snells_law_scalar(
        cos_incidence=direction,
        index_refraction=n,
        index_refraction_new=n_substrate,
    )

    cce = charge_collection_efficiency(
        absorption=absorption_substrate,
        thickness_implant=thickness_implant,
        cce_backsurface=cce_backsurface,
        cos_incidence=direction_substrate,
    )

    result = absorbance_substrate.average * cce

    return result


def probability_measurement(
    iqy: u.Quantity | na.AbstractScalar = 1 * u.electron / u.photon,
    cce: float | na.AbstractScalar = 1,
) -> na.AbstractScalar:
    r"""
    The probability that a photon absorbed in the epitaxial silicon layer results
    in at least one photoelectron measured by the sensor.

    For most of the electromagnetic spectrum, this quantity is nearly unity,
    but in the ultraviolet, there is a significant chance that all the
    photoelectrons associated with a photon recombine before being measured.

    Parameters
    ----------
    iqy
        The ideal quantum yield of the sensor in electrons per photon,
        calculated by :func:`quantum_yield_ideal`.
    cce
        The charge collection efficiency of the detector computed using
        :func:`charge_collection_efficiency`.

    Examples
    --------

    Plot the probability of measuring an absorbed photon vs the charge collection
    efficiency

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Define a grid of wavelengths
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Compute the ideal quantum yield of silicon for these wavelengths
        iqy = optika.sensors.quantum_yield_ideal(wavelength)

        # Compute the charge collection efficiency for each wavelength
        cce = optika.sensors.charge_collection_efficiency(
            absorption=optika.chemicals.Chemical("Si").absorption(wavelength),
        )

        # Compute the probability of measuring an absorbed photon
        # vs the charge collection efficiency
        p_m = optika.sensors.probability_measurement(iqy, cce)

        # Plot the results
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            na.plt.plot(
                wavelength,
                cce,
                ax=ax,
                label="charge collection efficiency",
            )
            na.plt.plot(
                wavelength,
                p_m,
                ax=ax,
                label="probability of measurement",
            )
            ax.set_xscale("log");
            ax.set_xlabel(f"wavelength ({ax.get_xlabel()})");
            ax.set_ylabel("probability");
            ax.legend();

    Notes
    -----

    The probability that `all` the electrons recombine before
    being measured is

    .. math::

        P_r = (1 - \text{CCE})^\text{IQY}

    Where :math:`\text{CCE}` is the charge collection efficiency,
    and :math:`\text{IQY}` is the ideal quantum yield of the sensor.
    So then the probability of a photon being measured is just the compliment
    of :math:`P_r`,

    .. math::

        P_m = 1 - P_r.
    """
    iqy = iqy.to(u.electron / u.photon).value
    p_r = (1 - cce) ** iqy
    p_m = 1 - p_r
    return p_m


def _discrete_gamma(
    mean: float | na.ScalarArray,
    vmr: float | na.ScalarArray,
    shape_random: None | dict[str, int] = None,
) -> na.ScalarArray:

    x = na.random.gamma(
        shape=mean / vmr,
        scale=vmr,
        shape_random=shape_random,
    )

    unit_x = x.unit
    if unit_x is not None:
        x = x.value

    x_frac, x_int = np.modf(x)
    x_frac = na.random.binomial(
        n=1,
        p=x_frac,
        shape_random=shape_random,
    )
    x = x_int + x_frac

    if unit_x is not None:
        x = x << unit_x

    return x


def electrons_measured(
    photons_absorbed: u.Quantity | na.AbstractScalar,
    absorption: u.Quantity | na.AbstractScalar,
    iqy: u.Quantity | na.AbstractScalar = 1 * u.electron / u.photon,
    thickness_implant: u.Quantity | na.AbstractScalar = _thickness_implant,
    cce_backsurface: u.Quantity | na.AbstractScalar = _cce_backsurface,
    fano_noise: u.Quantity | na.AbstractScalar = _fano_noise,
) -> na.AbstractScalar:
    r"""
    A random sample from the approximate distribution of measured electrons
    given the number of photons absorbed by the light-sensitive layer of the
    sensor.

    This function accounts for both Fano noise and recombination noise due to
    partial-charge collection.

    Parameters
    ----------
    photons_absorbed
        The number of photons absorbed by the light-sensitive layer of the sensor.
    absorption
        The absorption coefficient of the light-sensitive material for the
        wavelength of interest.
    iqy
        The ideal quantum yield of the sensor in electrons per photon.
    thickness_implant
        The thickness of the implant layer.
        Default is the value given in :cite:t:`Stern1994`.
    cce_backsurface
        The differential charge collection efficiency on the back surface
        of the sensor.
        Default is the value given in :cite:t:`Stern1994`.
    fano_noise
        The `Fano factor <https://en.wikipedia.org/wiki/Fano_factor>`_
        (ratio of the variance to the mean) of the Fano noise for this
        sensor material.
        Defaults to 0.1, the standard estimate of Fano noise in silicon
        :cite:p:`Janesick2001`.
        Must be in units of electrons per photon.

    Examples
    --------

    Plot the energy spectrum of 100 6 keV photons emitted from an Fe-55
    radioactive source.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Define the number of experiments to perform
        num_experiments = 100000

        # Define the expected number of photons
        # for each experiment
        photons_absorbed = na.broadcast_to(
            array=100* u.photon,
            shape=dict(experiment=num_experiments)
        ).astype(int)

        # Define the wavelength at which to sample the distribution
        wavelength = 5.9 * u.keV
        wavelength = wavelength.to(u.AA, equivalencies=u.spectral())

        # Compute the absorption coefficient of silicon at this wavelength
        absorption=optika.chemicals.Chemical("Si").absorption(wavelength)

        # Compute the ideal quantum yield of silicon for each wavelength
        iqy = optika.sensors.quantum_yield_ideal(wavelength)

        # Compute the actual number of electrons measured for each experiment
        electrons = optika.sensors.electrons_measured(
            photons_absorbed=photons_absorbed,
            absorption=absorption,
            iqy=iqy,
        )

        # Define the histogram bins
        step = 10
        bins = na.arange(
            electrons.value.min()-step/2,
            electrons.value.max()+step/2,
            step=step,
            axis="bin",
        ) * electrons.unit

        # Compute a histogram of resulting energy spectrum
        hist = na.histogram(
            electrons,
            bins=bins,
            axis="experiment",
        )

        # Plot the histogram
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            line = na.plt.stairs(
              hist.inputs,
              hist.outputs,
              ax=ax,
            )
    """
    shape = na.shape_broadcasted(
        photons_absorbed,
        absorption,
        absorbance,
        iqy,
        thickness_implant,
        cce_backsurface,
        fano_noise,
    )

    electrons_expected = iqy * photons_absorbed
    d = (2 / 12) * (u.electron / u.photon) ** 2
    f = fano_noise + d * (photons_absorbed - 1 * u.photon) / electrons_expected
    f = np.nan_to_num(f, neginf=fano_noise)

    a = absorption
    W = thickness_implant
    n0 = cce_backsurface
    aW = (a * W).to(u.dimensionless_unscaled).value

    fraction_complete = np.exp(-aW)
    fraction_partial = 1 - fraction_complete
    photons_absorbed_partial = na.random.binomial(
        n=photons_absorbed,
        p=fraction_partial,
        shape_random=shape,
    )
    photons_absorbed_complete = photons_absorbed - photons_absorbed_partial

    mean_p = n0 + (1 - n0) / (aW) + (1 - n0) / (1 - np.exp(aW))
    var_p = np.square(n0 - 1) * (4 / np.square(aW) - 1 / np.square(np.sinh(aW / 2))) / 4

    mean_n = iqy
    var_n = f * mean_n

    mean_p2 = np.square(mean_p)
    mean_n2 = np.square(mean_n)
    mean_i = mean_n * mean_p
    var_exp = (var_n * var_p) + (var_n * mean_p2) + (var_p * mean_n2)
    exp_var = mean_n * (mean_p - (var_p + mean_p2)) * u.electron / u.photon
    var_i = var_exp + exp_var

    mean_partial = photons_absorbed_partial * mean_i
    vmr_partial = var_i / mean_i * u.photon

    electrons_partial = _discrete_gamma(
        mean=mean_partial,
        vmr=np.maximum(vmr_partial - 1 / (6 * mean_partial) * u.electron**2, f * u.ph),
        shape_random=shape,
    )
    electrons_complete = _discrete_gamma(
        mean=photons_absorbed_complete * iqy,
        vmr=f * u.photon,
        shape_random=shape,
    )
    result = electrons_complete + electrons_partial

    return result


def signal(
    photons_expected: u.Quantity | na.AbstractScalar,
    absorption: u.Quantity | na.AbstractScalar,
    absorbance: float | na.AbstractScalar = 1,
    iqy: u.Quantity | na.AbstractScalar = 1 * u.electron / u.photon,
    thickness_implant: u.Quantity | na.AbstractScalar = _thickness_implant,
    cce_backsurface: u.Quantity | na.AbstractScalar = _cce_backsurface,
    fano_noise: u.Quantity | na.AbstractScalar = _fano_noise,
) -> na.AbstractScalar:
    r"""
    A random sample from the approximate distribution of measured electrons
    given the expected number of photons incident on the front surface of
    the sensor.

    This function adds shot noise to the expected number of photons,
    and then adds Fano noise and recombination noise using
    :func:`electrons_measured`.

    Parameters
    ----------
    photons_expected
        The `expected` number of photons incident on the detector surface.
    absorption
        The absorption coefficient of the light-sensitive material for the
        wavelength of interest.
    absorbance
        The fraction of incident energy absorbed by the light-sensitive layer
        of the detector computed using the average of :func:`absorbance`.
    iqy
        The ideal quantum yield of the sensor in electrons per photon.
    thickness_implant
        The thickness of the implant layer.
        Default is the value given in :cite:t:`Stern1994`.
    cce_backsurface
        The differential charge collection efficiency on the back surface
        of the sensor.
        Default is the value given in :cite:t:`Stern1994`.
    fano_noise
        The `Fano factor <https://en.wikipedia.org/wiki/Fano_factor>`_
        (ratio of the variance to the mean) of the Fano noise for this
        sensor material.
        Defaults to 0.1, the standard estimate of Fano noise in silicon
        :cite:p:`Janesick2001`.
        Must be in units of electrons per photon.

    Examples
    --------

    Plot the variance-to-mean ratio of the number of electrons measured by the sensor
    as a function of wavelength.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import named_arrays as na
        import optika

        # Define the number of experiments to perform
        num_experiments = 1000

        # Define the expected number of photons
        # for each experiment
        photons_expected = na.broadcast_to(
            array=100 * u.photon,
            shape=dict(experiment=num_experiments)
        )

        # Define a grid of wavelengths
        wavelength = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Compute the absorption coefficient of silicon at this wavelength
        absorption=optika.chemicals.Chemical("Si").absorption(wavelength)

        # Compute the fraction of light absorbed by the light-sensitive
        # region of the detector
        absorbance = optika.sensors.absorbance(wavelength).average

        # Compute the ideal quantum yield of silicon for each wavelength
        iqy = optika.sensors.quantum_yield_ideal(wavelength)

        # Compute the actual number of electrons measured for each experiment
        electrons = optika.sensors.signal(
            photons_expected=photons_expected,
            absorption=absorption,
            absorbance=absorbance,
            iqy=iqy,
        )

        # Plot the variance-to-mean ratio of the result
        # as a function of wavelength.
        fig, ax = plt.subplots(constrained_layout=True)
        na.plt.plot(
            wavelength,
            electrons.vmr("experiment"),
            ax=ax,
        );
        ax.set_xscale("log");
        ax.set_yscale("log");
        ax.set_xlabel(f"wavelength ({wavelength.unit:latex_inline})");
        ax.set_ylabel(f"variance-to-mean ratio ({electrons.unit:latex_inline})");
    """

    photons_absorbed_expected = absorbance * photons_expected.to(u.ph)
    photons_absorbed = na.random.poisson(photons_absorbed_expected).astype(int)

    return electrons_measured(
        photons_absorbed=photons_absorbed,
        absorption=absorption,
        iqy=iqy,
        thickness_implant=thickness_implant,
        cce_backsurface=cce_backsurface,
        fano_noise=fano_noise,
    )


@dataclasses.dataclass(eq=False, repr=False)
class AbstractImagingSensorMaterial(
    optika.materials.AbstractMaterial,
):
    """
    An interface representing the light-sensitive material of an imaging sensor.
    """

    @abc.abstractmethod
    def signal(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.rays.RayVectorArray:
        """
        Given a set of incident rays, compute the number of electrons
        measured by the sensor using :func:`signal`.

        Parameters
        ----------
        rays
            The rays incident on the sensor surface.
            The :attr:`optika.rays.RayVectorArray.intensity` field should
            either be in units of photons or energy.
        normal
            The vector perpendicular to the surface of the sensor.
        """

    @abc.abstractmethod
    def photons_incident(
        self,
        electrons: u.Quantity | na.AbstractScalar,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Given the number of electrons measured by the sensor,
        and a grid of wavelengths, compute the expected number of
        photons incident on the sensor.

        Parameters
        ----------
        electrons
            The number of electrons measured by the sensor.
        wavelength
            An assumed grid of wavelengths for the incident photons.
        direction
            An assumed propagation direction for the incident photons.
        normal
            The vector perpendicular to the surface of the sensor.
        """

    @abc.abstractmethod
    def charge_diffusion(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.rays.RayVectorArray:
        """
        Given a set of incident rays, compute how much the position of each ray
        is perturbed due to charge diffusion within the sensor.

        Parameters
        ----------
        rays
            The rays incident on the sensor surface.
        normal
            The vector perpendicular to the surface of the sensor.
        """


@dataclasses.dataclass(eq=False, repr=False)
class IdealImagingSensorMaterial(
    optika.materials.Vacuum,
    AbstractImagingSensorMaterial,
):
    """
    An idealized sensor material with a quantum efficiency of unity
    and no charge diffusion.
    """

    def signal(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.rays.RayVectorArray:

        intensity = rays.intensity
        if not intensity.unit.is_equivalent(u.photon):
            h = astropy.constants.h
            c = astropy.constants.c
            intensity = intensity / (h * c / rays.wavelength) * u.photon
        electrons = intensity * u.electron / u.photon
        electrons = electrons.to(u.electron)

        result = dataclasses.replace(rays, intensity=electrons)

        return result

    def photons_incident(
        self,
        electrons: u.Quantity | na.AbstractScalar,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        return electrons * u.photon / u.electron

    def charge_diffusion(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.rays.RayVectorArray:
        return rays


@dataclasses.dataclass(eq=False, repr=False)
class AbstractCCDMaterial(
    AbstractImagingSensorMaterial,
):
    """
    An interface representing the light-sensitive material of a CCD sensor.
    """

    @property
    def transformation(self) -> None:
        return None

    @functools.cached_property
    def _chemical(self) -> optika.chemicals.Chemical:
        return optika.chemicals.Chemical("Si")

    @functools.cached_property
    def _chemical_oxide(self) -> optika.chemicals.Chemical:
        return optika.chemicals.Chemical("SiO2_llnl_cxro_rodriguez")

    @property
    def fano_noise(self) -> u.Quantity:
        """
        The `Fano factor <https://en.wikipedia.org/wiki/Fano_factor>`_
        (ratio of the variance to the mean) of the Fano noise for this
        sensor material.

        This parameter has some variation in the literature,
        this implementation uses the value given by :cite:t:`Rodrigues2021`.
        """
        return 0.119 * u.electron / u.photon

    def index_refraction(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        return rays.index_refraction

    def attenuation(
        self,
        rays: optika.rays.AbstractRayVectorArray,
    ) -> na.ScalarLike:
        return rays.attenuation

    @property
    def is_mirror(self) -> bool:
        return False


@dataclasses.dataclass(eq=False, repr=False)
class AbstractBackilluminatedCCDMaterial(
    AbstractCCDMaterial,
):
    """
    An interface representing the light-sensitive material of a backilluminated
    CCD sensor.
    """

    @property
    @abc.abstractmethod
    def thickness_oxide(self) -> u.Quantity | na.AbstractScalar:
        """
        The thickness of the oxide layer on the back surface of the CCD sensor.
        """

    @property
    @abc.abstractmethod
    def thickness_implant(self) -> u.Quantity | na.AbstractScalar:
        """
        The thickness of the implant layer of the CCD sensor.
        """

    @property
    @abc.abstractmethod
    def thickness_substrate(self) -> u.Quantity | na.AbstractScalar:
        """the thickness of the entire CCD silicon substrate"""

    @property
    @abc.abstractmethod
    def roughness_oxide(self):
        """The RMS roughness of the oxide layer surface."""

    @property
    @abc.abstractmethod
    def roughness_substrate(self):
        """The RMS roughness of the substrate surface."""

    @property
    @abc.abstractmethod
    def cce_backsurface(self) -> float | na.AbstractScalar:
        """
        The charge collection efficiency on the backsurface of the CCD sensor.
        """

    @property
    @abc.abstractmethod
    def depletion(self) -> AbstractDepletionModel:
        """
        A model of the sensor's depletion region.
        """

    def width_charge_diffusion(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        The standard deviation of the charge diffusion kernel for this sensor.
        Calculated using :func:`optika.sensors.charge_diffusion`.

        Parameters
        ----------
        rays
            The rays incident on the sensor surface.
        normal
            The vector perpendicular to the surface of the sensor.
        """
        return optika.sensors.charge_diffusion(
            self._chemical.absorption(rays.wavelength),
            thickness_substrate=self.thickness_substrate,
            thickness_depletion=self.depletion.thickness,
        )

    def quantum_yield_ideal(
        self,
        wavelength: u.Quantity | na.AbstractScalar,
    ) -> u.Quantity | na.AbstractScalar:
        """
        Compute the ideal quantum yield of this CCD sensor material using
        :func:`optika.sensors.quantum_yield_ideal`.

        Parameters
        ----------
        wavelength
            The wavelength of the incident light
        """
        return quantum_yield_ideal(wavelength)

    def absorbance(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.vectors.PolarizationVectorArray:
        """
        Compute the fraction of energy absorbed by the light-sensitive region
        of the sensor.

        Parameters
        ----------
        rays
            The light rays incident on the CCD surface.
        normal
            The vector perpendicular to the surface of the CCD sensor.
        """
        return absorbance(
            wavelength=rays.wavelength,
            direction=-rays.direction @ normal,
            n=rays.index_refraction,
            thickness_oxide=self.thickness_oxide,
            thickness_substrate=self.thickness_substrate,
            chemical_oxide=self._chemical_oxide,
            chemical_substrate=self._chemical,
            roughness_oxide=self.roughness_oxide,
            roughness_substrate=self.roughness_substrate,
        )

    def charge_collection_efficiency(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Compute the charge collection efficiency of this CCD sensor material
        using :func:`charge_collection_efficiency`.

        Parameters
        ----------
        rays
            The light rays incident on the CCD surface.
        normal
            The vector perpendicular to the surface of the CCD sensor.
        """
        return charge_collection_efficiency(
            absorption=self._chemical.absorption(rays.wavelength),
            thickness_implant=self.thickness_implant,
            cce_backsurface=self.cce_backsurface,
            cos_incidence=-rays.direction @ normal,
        )

    def quantum_efficiency_effective(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Compute the effective quantum efficiency of this CCD material using
        :func:`optika.sensors.quantum_efficiency_effective`.

        Parameters
        ----------
        rays
            The light rays incident on the CCD surface
        normal
            The vector perpendicular to the surface of the CCD.
        """
        k = rays.attenuation * rays.wavelength / (4 * np.pi)
        n = rays.index_refraction + k * 1j

        return quantum_efficiency_effective(
            wavelength=rays.wavelength,
            direction=rays.direction,
            n=n,
            thickness_oxide=self.thickness_oxide,
            thickness_implant=self.thickness_implant,
            thickness_substrate=self.thickness_substrate,
            cce_backsurface=self.cce_backsurface,
            chemical_oxide=self._chemical_oxide,
            chemical_substrate=self._chemical,
            roughness_oxide=self.roughness_oxide,
            roughness_substrate=self.roughness_substrate,
            normal=normal,
        )

    def quantum_efficiency(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Compute the quantum efficiency of this CCD material using
        :meth:`quantum_efficiency_effective` and
        :meth:`quantum_yield_ideal`.

        Parameters
        ----------
        rays
            The light rays incident on the CCD surface
        normal
            The vector perpendicular to the surface of the CCD.
        """
        iqy = self.quantum_yield_ideal(rays.wavelength)
        eqe = self.quantum_efficiency_effective(rays, normal)
        return iqy * eqe

    def probability_measurement(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Compute the probability of measuring an absorbed photon for this sensor
        using :func:`probability_measurement`.

        Parameters
        ----------
        rays
            The light rays incident on the CCD surface
        normal
            The vector perpendicular to the surface of the CCD.
        """
        return probability_measurement(
            iqy=self.quantum_yield_ideal(rays.wavelength),
            cce=self.charge_collection_efficiency(rays, normal),
        )

    def electrons_measured(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.rays.RayVectorArray:

        intensity = rays.intensity
        wavelength = rays.wavelength

        electrons = electrons_measured(
            photons_absorbed=intensity,
            absorption=self._chemical.absorption(wavelength),
            iqy=self.quantum_yield_ideal(wavelength),
            thickness_implant=self.thickness_implant,
            cce_backsurface=self.cce_backsurface,
            fano_noise=self.fano_noise,
        )

        result = dataclasses.replace(rays, intensity=electrons)

        return result

    def signal(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.rays.RayVectorArray:

        intensity = rays.intensity
        wavelength = rays.wavelength

        if not intensity.unit.is_equivalent(u.photon):
            h = astropy.constants.h
            c = astropy.constants.c
            intensity = intensity / (h * c / wavelength) * u.photon

        electrons = signal(
            photons_expected=intensity,
            absorption=self._chemical.absorption(wavelength),
            absorbance=self.absorbance(rays, normal).average,
            iqy=self.quantum_yield_ideal(wavelength),
            thickness_implant=self.thickness_implant,
            cce_backsurface=self.cce_backsurface,
            fano_noise=self.fano_noise,
        )

        result = dataclasses.replace(rays, intensity=electrons)

        return result

    def photons_incident(
        self,
        electrons: u.Quantity | na.AbstractScalar,
        wavelength: u.Quantity | na.AbstractScalar,
        direction: na.AbstractCartesian3dVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.AbstractScalar:
        """
        Compute the expected number of incident photons for a given number
        of electrons.

        Parameters
        ----------
        electrons
            The energy collected by the sensor in units of electrons.
        wavelength
            The assumed wavelength of the incident photons.
        direction
            The assumed direction of the incident photons.
        normal
            The vector perpendicular to the surface of the sensor.
        """

        qe = self.quantum_efficiency(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength,
                direction=direction,
            ),
            normal=normal,
        )

        return electrons / qe

    def charge_diffusion(
        self,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> optika.rays.RayVectorArray:

        width = self.width_charge_diffusion(rays, normal)

        position = dataclasses.replace(
            rays.position,
            x=na.random.normal(rays.position.x, width),
            y=na.random.normal(rays.position.y, width),
        )

        rays = dataclasses.replace(
            rays,
            position=position,
        )

        return rays

    def efficiency(
        self,
        rays: optika.rays.AbstractRayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ) -> na.ScalarLike:
        return 1


@dataclasses.dataclass(eq=False, repr=False)
class AbstractStern1994BackilluminatedCCDMaterial(
    AbstractBackilluminatedCCDMaterial,
):
    """
    A CCD material that is uses the method of :cite:t:`Stern1994` to compute
    the quantum efficiency.
    """

    @property
    @abc.abstractmethod
    def quantum_efficiency_measured(self) -> na.FunctionArray:
        """
        The measured quantum efficiency that will be fit by the function
        :func:`optika.sensors.quantum_efficiency_effective`.
        """

    @functools.cached_property
    def _quantum_efficiency_fit(self) -> dict[str, float | u.Quantity]:
        qe_measured = self.quantum_efficiency_measured

        unit_thickness_oxide = u.AA
        unit_thickness_implant = u.AA
        unit_roughness = u.nm
        unit_cce_backsurface = u.dimensionless_unscaled

        def eqe_rms_difference(x: tuple[float, float, float, float]):
            (
                thickness_oxide,
                thickness_implant,
                roughness_substrate,
                cce_backsurface,
            ) = x
            qe_fit = quantum_efficiency_effective(
                wavelength=qe_measured.inputs,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
                thickness_oxide=thickness_oxide << unit_thickness_oxide,
                thickness_implant=thickness_implant << unit_thickness_implant,
                thickness_substrate=self.thickness_substrate,
                roughness_oxide=self.roughness_oxide,
                roughness_substrate=roughness_substrate << unit_roughness,
                cce_backsurface=cce_backsurface << unit_cce_backsurface,
            )

            return np.sqrt(np.mean(np.square(qe_measured.outputs - qe_fit))).ndarray

        thickness_oxide_guess = 50 * u.AA
        thickness_implant_guess = 2317 * u.AA
        roughness_substrate_guess = 5 * u.nm
        cce_backsurface_guess = 0.21 * u.dimensionless_unscaled

        fit = scipy.optimize.minimize(
            fun=eqe_rms_difference,
            x0=[
                thickness_oxide_guess.to_value(unit_thickness_oxide),
                thickness_implant_guess.to_value(unit_thickness_implant),
                roughness_substrate_guess.to_value(unit_roughness),
                cce_backsurface_guess.to_value(unit_cce_backsurface),
            ],
            bounds=[
                (0, None),
                (0, None),
                (0, None),
                (0, 1),
            ],
            method="nelder-mead",
        )

        (
            thickness_oxide,
            thickness_implant,
            roughness_substrate,
            cce_backsurface,
        ) = fit.x

        thickness_oxide = thickness_oxide << unit_thickness_oxide
        thickness_implant = thickness_implant << unit_thickness_implant
        roughness_substrate = roughness_substrate << unit_roughness
        cce_backsurface = cce_backsurface << unit_cce_backsurface

        return dict(
            thickness_oxide=thickness_oxide,
            thickness_implant=thickness_implant,
            roughness_substrate=roughness_substrate,
            cce_backsurface=cce_backsurface,
        )

    @property
    def thickness_oxide(self) -> u.Quantity:
        return self._quantum_efficiency_fit["thickness_oxide"]

    @property
    def thickness_implant(self) -> u.Quantity:
        return self._quantum_efficiency_fit["thickness_implant"]

    @property
    def roughness_oxide(self):
        return 0 * u.nm

    @property
    def roughness_substrate(self):
        return self._quantum_efficiency_fit["roughness_substrate"]

    @property
    def cce_backsurface(self) -> float:
        return self._quantum_efficiency_fit["cce_backsurface"]



import pathlib
import numpy as np
import astropy.units as u
import named_arrays as na
from .._depletion import E2VCCD64ThickDepletionModel
from .._materials import AbstractStern1994BackilluminatedCCDMaterial

__all__ = [
    "E2VCCD203Material",
]


class E2VCCD203Material(
    AbstractStern1994BackilluminatedCCDMaterial,
):
    """
    A model of the light-sensitive material of the custom e2v CCD sensors
    on board the Atmospheric Imaging Assembly :cite:p:`Lemen2012` from
    :cite:t:`Boerner2012`

    This model uses the :class:`optika.sensors.E2VCCD64ThickDepletionModel`
    to represent the depletion region.

    Examples
    --------

    Plot the measured AIA CCD quantum efficiency vs the fitted
    quantum efficiency calculated using the method of :cite:t:`Stern1994`.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Create a new instance of the e2v CCD97 light-sensitive material
        material = optika.sensors.E2VCCD203Material()

        # Store the wavelengths at which the QE was measured
        wavelength_measured = material.quantum_efficiency_measured.inputs

        # Store the QE measurements
        qe_measured = material.quantum_efficiency_measured.outputs

        # Define a grid of wavelengths with which to evaluate the fitted QE
        wavelength_fit = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Evaluate the fitted QE using the given wavelengths
        qe_fit = material.quantum_efficiency_effective(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the measured QE vs the fitted QE
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            na.plt.scatter(
                wavelength_measured,
                qe_measured,
                label="measured",
            )
            na.plt.plot(
                wavelength_fit,
                qe_fit,
                label="fit",
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({wavelength_fit.unit:latex_inline})")
            ax.set_ylabel("quantum efficiency")
            ax.legend()

    The thickness of the oxide layer found by the fit is

    .. jupyter-execute::

        material.thickness_oxide

    The thickness of the implant layer found by the fit is

    .. jupyter-execute::

        material.thickness_implant

    The thickness of the substrate is

    .. jupyter-execute::

        material.thickness_substrate

    The differential charge collection efficiency at the backsurface
    found by the fit is

    .. jupyter-execute::

        material.cce_backsurface

    And the roughness of the substrate found by the fit is

    .. jupyter-execute::

        material.roughness_substrate

    |

    Plot the width of the charge diffusion kernel for this sensor as a function
    of wavelength.

    .. jupyter-execute::

        # Compute the width of the charge diffusion kernel
        # for each wavelength.
        width = material.width_charge_diffusion(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the results
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            na.plt.plot(
                wavelength_fit,
                width,
                ax=ax,
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({ax.get_xlabel()})")
            ax.set_ylabel(f"width ({ax.get_ylabel()})")
    """

    @property
    def quantum_efficiency_measured(self) -> na.FunctionArray:
        directory = pathlib.Path(__file__).parent
        (
            wavelength_1,
            qe_1,
            wavelength_2,
            qe_2,
            wavelength_3,
            qe_3,
            wavelength_4,
            qe_4,
        ) = np.genfromtxt(
            fname=directory / "e2v_ccd203_qe_boerner2012.csv",
            skip_header=2,
            delimiter=",",
            unpack=True,
        )
        wavelength = (wavelength_1 + wavelength_2 + wavelength_3 + wavelength_4) / 4
        wavelength = wavelength << u.AA
        qe = (qe_1 + qe_2 + qe_3 + qe_4) / 4
        return na.FunctionArray(
            inputs=na.ScalarArray(wavelength, axes="wavelength"),
            outputs=na.ScalarArray(qe, axes="wavelength"),
        )

    @property
    def thickness_substrate(self) -> u.Quantity:
        return 16 * u.um

    @property
    def depletion(self) -> E2VCCD64ThickDepletionModel:
        return E2VCCD64ThickDepletionModel()

    @property
    def shape(self) -> dict[str, int]:
        return dict()



import astropy.units as u
import named_arrays as na
from .._depletion import E2VCCD64ThinDepletionModel
from .._materials import AbstractStern1994BackilluminatedCCDMaterial

__all__ = [
    "TektronixTK512CBMaterial",
]


class TektronixTK512CBMaterial(
    AbstractStern1994BackilluminatedCCDMaterial,
):
    """
    A model of the light-sensitive material of a Tektronix TK512CB sensor based on
    measurements by :cite:t:`Stern1994`.

    This model uses the :class:`optika.sensors.E2VCCD64ThinDepletionModel`
    to represent the depletion region.

    Examples
    --------

    Plot the measured TK512CB quantum efficiency vs the fitted
    quantum efficiency calculated using the method of :cite:t:`Stern1994`.

    .. jupyter-execute::

        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # Create a new instance of the e2v CCD97 light-sensitive material
        material = optika.sensors.TektronixTK512CBMaterial()

        # Store the wavelengths at which the QE was measured
        wavelength_measured = material.quantum_efficiency_measured.inputs

        # Store the QE measurements
        qe_measured = material.quantum_efficiency_measured.outputs

        # Define a grid of wavelengths with which to evaluate the fitted QE
        wavelength_fit = na.geomspace(10, 10000, axis="wavelength", num=1001) * u.AA

        # Evaluate the fitted QE using the given wavelengths
        qe_fit = material.quantum_efficiency_effective(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the measured QE vs the fitted QE
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            na.plt.scatter(
                wavelength_measured,
                qe_measured,
                label="measured",
            )
            na.plt.plot(
                wavelength_fit,
                qe_fit,
                label="fit",
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({wavelength_fit.unit:latex_inline})")
            ax.set_ylabel("quantum efficiency")
            ax.legend()

    The thickness of the oxide layer found by the fit is

    .. jupyter-execute::

        material.thickness_oxide

    The thickness of the implant layer found by the fit is

    .. jupyter-execute::

        material.thickness_implant

    The thickness of the substrate is

    .. jupyter-execute::

        material.thickness_substrate

    The differential charge collection efficiency at the backsurface
    found by the fit is

    .. jupyter-execute::

        material.cce_backsurface

    And the roughness of the substrate found by the fit is

    .. jupyter-execute::

        material.roughness_substrate

    |

    Plot the width of the charge diffusion kernel for this sensor as a function
    of wavelength.

    .. jupyter-execute::

        # Compute the width of the charge diffusion kernel
        # for each wavelength.
        width = material.width_charge_diffusion(
            rays=optika.rays.RayVectorArray(
                wavelength=wavelength_fit,
                direction=na.Cartesian3dVectorArray(0, 0, 1),
            ),
            normal=na.Cartesian3dVectorArray(0, 0, -1),
        )

        # Plot the results
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots()
            na.plt.plot(
                wavelength_fit,
                width,
                ax=ax,
            )
            ax.set_xscale("log")
            ax.set_xlabel(f"wavelength ({ax.get_xlabel()})")
            ax.set_ylabel(f"width ({ax.get_ylabel()})")
    """

    @property
    def quantum_efficiency_measured(self) -> na.FunctionArray:
        wavelength = [
            13.3,
            23.6,
            44.7,
            67.6,
            114.0,
            135.5,
            171.4,
            256.0,
            303.8,
            461.0,
            584.0,
            736.0,
            1215.5,
            2537.0,
            3650.0,
            4050.0,
        ] * u.AA
        qe = [
            0.91,
            0.80,
            0.48,
            0.32,
            0.42,
            0.86,
            0.82,
            0.60,
            0.58,
            0.53,
            0.30,
            0.085,
            0.055,
            0.06,
            0.09,
            0.29,
        ] * u.dimensionless_unscaled
        return na.FunctionArray(
            inputs=na.ScalarArray(wavelength, axes="wavelength"),
            outputs=na.ScalarArray(qe, axes="wavelength"),
        )

    @property
    def thickness_substrate(self) -> u.Quantity:
        return 7 * u.um

    @property
    def depletion(self) -> E2VCCD64ThinDepletionModel:
        return E2VCCD64ThinDepletionModel()

    @property
    def shape(self) -> dict[str, int]:
        return dict()



"""
A collection of tools for simulating measuring and tolerancing of optical surfaces.
"""

from ._slope_error import AbstractSlopeErrorParameters, SlopeErrorParameters
from ._roughness import AbstractRoughnessParameters, RoughnessParameters

__all__ = [
    "AbstractSlopeErrorParameters",
    "SlopeErrorParameters",
    "AbstractRoughnessParameters",
    "RoughnessParameters",
]



"""
Optical systems consisting of multiple optical surfaces.
"""

from __future__ import annotations
from typing import Sequence, Callable, Any, ClassVar
import abc
import dataclasses
import functools
import astropy.units as u
import numpy as np
import numpy.typing as npt
import matplotlib.axes
import named_arrays as na
import optika

__all__ = [
    "AbstractSystem",
    "AbstractSequentialSystem",
    "SequentialSystem",
]


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSystem(
    optika.mixins.Plottable,
    optika.mixins.Printable,
    optika.mixins.Transformable,
    optika.mixins.Shaped,
):
    """
    An interface describing an optical system.

    Could potentially be sequential or non-sequential.
    """

    @abc.abstractmethod
    def image(
        self,
        scene: na.FunctionArray[na.SpectralPositionalVectorArray, na.AbstractScalar],
        **kwargs: Any,
    ) -> na.SpectralPositionalVectorArray:
        """
        Forward model of the optical system.
        Maps the given spectral radiance of a scene to detector counts.

        Parameters
        ----------
        scene
            The spectral radiance of the scene as a function of wavelength
            and field position.
        kwargs
            Additional keyword arguments used by subclass implementations
            of this method.
        """


@dataclasses.dataclass(eq=False, repr=False)
class AbstractSequentialSystem(
    AbstractSystem,
):
    """
    An interface describing a sequential optical system.

    A sequential optical system is a system where the ordering of the optical
    surfaces is known in advance.
    """

    @property
    @abc.abstractmethod
    def object(self) -> optika.surfaces.AbstractSurface:
        """
        The external object being imaged or illuminated by this system.

        If :obj:`None`, the external object is assumed to be at infinity.
        """

    @property
    def object_is_at_infinity(self) -> bool:
        """
        A boolean flag indicating if the object is at infinity.

        If :attr:`object` doesn't have an aperture,
        it is assumed that the object is at infinity.
        """
        aperture_obj = self.object.aperture
        if aperture_obj is not None:
            unit_obj = na.unit_normalized(aperture_obj.bound_lower)
            if unit_obj.is_equivalent(u.mm):
                result = False
            elif unit_obj.is_equivalent(u.dimensionless_unscaled):
                result = True
            else:  # pragma: nocover
                raise ValueError(f"Unrecognized unit for object aperture, {unit_obj}")
        else:
            result = True
        return result

    @property
    @abc.abstractmethod
    def surfaces(self) -> Sequence[optika.surfaces.AbstractSurface]:
        """
        a sequence of surfaces representing this optical system.

        At least one of these surfaces needs to be marked as the pupil surface,
        and if the object surface is not marked as the field stop, one of these
        surfaces needs to be marked as the field stop.
        """

    @property
    @abc.abstractmethod
    def sensor(self) -> None | optika.sensors.AbstractImagingSensor:
        """
        The imaging sensor that measures the light captured by this system.

        This is the last surface in the optical system.
        """

    @property
    @abc.abstractmethod
    def axis_surface(self) -> str:
        """
        The name of the logical axis representing the sequence of surfaces.
        """

    @property
    def surfaces_all(self) -> list[optika.surfaces.AbstractSurface]:
        """
        Concatenate :attr:`object` with :attr:`surfaces` into a single list of
        surfaces.
        """
        obj = self.object
        if obj is not None:
            result = [obj]
        else:
            result = []

        result += list(self.surfaces)

        sensor = self.sensor
        if sensor is not None:
            result += [sensor]

        if not any(s.is_field_stop for s in result):
            result[0] = dataclasses.replace(result[0], is_field_stop=True)
        return result

    @property
    @abc.abstractmethod
    def grid_input(self) -> optika.vectors.ObjectVectorArray:
        """
        The input grid to sample with rays.

        This grid is simultaneously projected onto both the field stop and the
        pupil stop.

        Positions on the stop can be specified in either absolute or normalized
        coordinates. Using normalized coordinates allows for injecting different
        grid types (cylindrical, stratified random, etc.) without specifying the
        scale of the stop surface.

        If positions are specified in absolute units, they are measured in the
        coordinate system of the corresponding stop surface.
        """

    @property
    def _indices_field_stop(self) -> list[int]:
        return [i for i, s in enumerate(self.surfaces_all) if s.is_field_stop]

    @property
    def _indices_pupil_stop(self) -> list[int]:
        return [i for i, s in enumerate(self.surfaces_all) if s.is_pupil_stop]

    @property
    def index_field_stop(self) -> int:
        """
        The index of the field stop in :attr:`surfaces_all`.
        """
        indices = self._indices_field_stop
        if not indices:
            raise ValueError(
                "Field stop is not defined for this system."
                "Set `is_field_stop=True` for at least one surface in this system."
            )
        return indices[~0]

    @property
    def index_pupil_stop(self) -> int:
        """
        The index of the pupil stop in :attr:`surfaces_all`.
        """
        indices = self._indices_pupil_stop
        if not indices:
            raise ValueError(
                "Pupil stop is not defined for this system."
                "Set `is_pupil_stop=True` for at least one surface in this "
                "system."
            )
        return indices[~0]

    @property
    def field_stop(self) -> optika.surfaces.AbstractSurface:
        """
        The field stop surface.
        """
        return self.surfaces_all[self.index_field_stop]

    @property
    def pupil_stop(self) -> optika.surfaces.AbstractSurface:
        """
        The pupil stop surface.
        """
        return self.surfaces_all[self.index_pupil_stop]

    @classmethod
    def _ray_error(
        cls,
        a: na.Cartesian2dVectorArray,
        rays: optika.rays.RayVectorArray,
        subsystem: list[optika.surfaces.AbstractSurface],
        grid_last: na.Cartesian2dVectorArray,
        component_variable: str,
        component_target: str,
        zfunc: Callable[[na.Cartesian2dVectorArray], na.ScalarLike],
    ):
        rays_component_variable = getattr(rays, component_variable)
        rays_component_variable.x = a.x
        rays_component_variable.y = a.y
        rays_component_variable.z = zfunc(a)

        rays = optika.propagators.propagate_rays(
            propagators=subsystem[1:],
            rays=rays,
        )

        transformation_last = subsystem[~0].transformation
        if transformation_last is not None:
            rays = transformation_last.inverse(rays)

        rays_component_target = getattr(rays, component_target)
        grid_last_trial = na.Cartesian2dVectorArray(
            x=rays_component_target.x,
            y=rays_component_target.y,
        )
        result = grid_last_trial - grid_last
        return result

    def _calc_rayfunction_stops_only(
        self,
        wavelength_input: na.ScalarLike,
        axis_pupil_stop: str,
        axis_field_stop: str,
        samples_pupil_stop: int = 101,
        samples_field_stop: int = 101,
    ) -> optika.rays.RayFunctionArray:
        result = optika.rays.RayFunctionArray(
            inputs=optika.vectors.ObjectVectorArray(
                wavelength=wavelength_input,
            ),
            outputs=optika.rays.RayVectorArray(
                wavelength=wavelength_input,
            ),
        )

        surfaces = self.surfaces_all

        indices_pupil_stop = self._indices_pupil_stop
        indices_field_stop = self._indices_field_stop

        if not indices_pupil_stop:
            raise ValueError("pupil not defined")
        if not indices_field_stop:
            raise ValueError("field stop not defined")

        while indices_pupil_stop and indices_field_stop:
            index_pupil_stop = indices_pupil_stop[0]
            index_field_stop = indices_field_stop[0]

            index_first = min(index_pupil_stop, index_field_stop)
            index_last = max(index_pupil_stop, index_field_stop)

            subsystem = surfaces[index_first : index_last + 1]

            surface_first = subsystem[0]
            surface_last = subsystem[~0]

            aperture_first = surface_first.aperture
            aperture_last = surface_last.aperture

            grid_first = np.moveaxis(
                a=aperture_first.wire(num=samples_pupil_stop),
                source="wire",
                destination=axis_pupil_stop,
            )
            grid_first = na.Cartesian2dVectorArray(grid_first.x, grid_first.y)

            grid_last = np.moveaxis(
                a=aperture_last.wire(num=samples_field_stop),
                source="wire",
                destination=axis_field_stop,
            )
            grid_last = na.Cartesian2dVectorArray(grid_last.x, grid_last.y)

            if surface_first.is_pupil_stop:
                indices_pupil_stop.pop(0)
                result.inputs.pupil = grid_first
                result.inputs.field = grid_last
            else:
                indices_field_stop.pop(0)
                result.inputs.field = grid_first
                result.inputs.pupil = grid_last

            if na.unit(grid_first).is_equivalent(u.mm):
                result.outputs.position = na.Cartesian3dVectorArray(
                    x=grid_first.x,
                    y=grid_first.y,
                    z=surface_first.sag(grid_first),
                )
                result.outputs.direction = na.Cartesian3dVectorArray(0, 0, 1)
                component_variable = "direction"

                def zfunc(xy: na.AbstractCartesian3dVectorArray):
                    return np.sqrt(1 - np.square(xy.length))

            elif na.unit(grid_first).is_equivalent(u.dimensionless_unscaled):
                result.outputs.direction = na.Cartesian3dVectorArray(
                    x=grid_first.x,
                    y=grid_first.y,
                    z=np.sqrt(1 - np.square(grid_first.length)),
                )
                result.outputs.position = na.Cartesian3dVectorArray() * u.mm
                component_variable = "position"

                def zfunc(xy: na.AbstractCartesian3dVectorArray):
                    return surface_first.sag(xy)

            else:
                raise ValueError(f"unrecognized input grid unit, {na.unit(grid_first)}")

            if surface_first.transformation is not None:
                result.outputs = surface_first.transformation(result.outputs)

            if na.unit(grid_last).is_equivalent(u.mm):
                component_target = "position"
            elif na.unit(grid_last).is_equivalent(u.dimensionless_unscaled):
                component_target = "direction"
            else:
                raise ValueError(f"unrecognized output grid unit, {na.unit(grid_last)}")

            variables = getattr(result.outputs, component_variable)

            root = na.optimize.root_newton(
                function=functools.partial(
                    self._ray_error,
                    rays=result.outputs,
                    subsystem=subsystem,
                    grid_last=grid_last,
                    component_variable=component_variable,
                    component_target=component_target,
                    zfunc=zfunc,
                ),
                guess=na.Cartesian2dVectorArray(
                    x=variables.x,
                    y=variables.y,
                ),
            )

            variables.x = root.x
            variables.y = root.y
            variables.z = zfunc(root)

        return result

    def _calc_rayfunction_stops(
        self,
        wavelength_input: na.ScalarLike,
        axis_pupil_stop: str,
        axis_field_stop: str,
        samples_pupil_stop: int = 101,
        samples_field_stop: int = 101,
    ) -> optika.rays.RayFunctionArray:
        surfaces = self.surfaces_all

        index_pupil_stop = self.index_pupil_stop
        index_field_stop = self.index_field_stop

        index_stop = min(index_pupil_stop, index_field_stop)

        subsystem = surfaces[index_stop::-1]

        rays_stop = self._calc_rayfunction_stops_only(
            wavelength_input=wavelength_input,
            axis_pupil_stop=axis_pupil_stop,
            axis_field_stop=axis_field_stop,
            samples_pupil_stop=samples_pupil_stop,
            samples_field_stop=samples_field_stop,
        )

        result = rays_stop.copy_shallow()
        result.outputs = optika.propagators.propagate_rays(
            propagators=subsystem,
            rays=rays_stop.outputs,
        )

        obj = subsystem[~0]
        rays = result.outputs
        if obj.transformation is not None:
            rays = obj.transformation.inverse(rays)

        where = rays.direction @ obj.sag.normal(rays.position) > 0
        result.outputs.direction[where] = -result.outputs.direction[where]

        if self.transformation is not None:
            result.outputs = self.transformation(result.outputs)

        return result

    _axis_pupil_stop: ClassVar[str] = "_stop_pupil"
    _axis_field_stop: ClassVar[str] = "_stop_field"

    @functools.cached_property
    def rayfunction_stops(self) -> optika.rays.RayFunctionArray:
        """
        A rayfunction defined on the input surface of the optical system,
        which is designed to exactly strike the borders of both the field
        stop and the pupil stop.
        """
        return self._calc_rayfunction_stops(
            wavelength_input=self.grid_input.wavelength,
            axis_pupil_stop=self._axis_pupil_stop,
            axis_field_stop=self._axis_field_stop,
            samples_pupil_stop=21,
            samples_field_stop=21,
        )

    @property
    def field_min(self) -> na.AbstractCartesian2dVectorArray:
        """
        The lower left corner of this optical system's field of view.
        """
        axis = (self._axis_field_stop, self._axis_pupil_stop)
        if self.object_is_at_infinity:
            angles = optika.angles(self.rayfunction_stops.outputs.direction)
            return angles.min(axis)
        else:
            return self.rayfunction_stops.outputs.position.xy.min(axis)

    @property
    def field_max(self) -> na.AbstractCartesian2dVectorArray:
        """
        The upper right corner of this optical system's field of view.
        """
        axis = (self._axis_field_stop, self._axis_pupil_stop)
        if self.object_is_at_infinity:
            angles = optika.angles(self.rayfunction_stops.outputs.direction)
            return angles.max(axis)
        else:
            return self.rayfunction_stops.outputs.position.xy.max(axis)

    @property
    def pupil_min(self) -> na.AbstractCartesian2dVectorArray:
        """
        The lower left corner of this optical system's entrance pupil in
        physical units.
        """
        axis = (self._axis_field_stop, self._axis_pupil_stop)
        if self.object_is_at_infinity:
            return self.rayfunction_stops.outputs.position.xy.min(axis)
        else:
            angles = optika.angles(self.rayfunction_stops.outputs.direction)
            return angles.min(axis)

    @property
    def pupil_max(self):
        """
        The upper right corner of this optical system's entrance pupil in
        physical units.
        """
        axis = (self._axis_field_stop, self._axis_pupil_stop)
        if self.object_is_at_infinity:
            return self.rayfunction_stops.outputs.position.xy.max(axis)
        else:
            angles = optika.angles(self.rayfunction_stops.outputs.direction)
            return angles.max(axis)

    def _denormalize_grid(
        self,
        grid: optika.vectors.ObjectVectorArray,
        normalized_field: bool = True,
        normalized_pupil: bool = True,
    ) -> optika.vectors.ObjectVectorArray:

        if (not normalized_field) and (not normalized_pupil):
            return grid

        axis_field = self._axis_field_stop
        axis_pupil = self._axis_pupil_stop

        rayfunction_stops = self._calc_rayfunction_stops(
            wavelength_input=grid.wavelength,
            axis_pupil_stop=axis_pupil,
            axis_field_stop=axis_field,
            samples_pupil_stop=21,
            samples_field_stop=21,
        )

        result = grid.copy_shallow()

        object_is_at_infinity = self.object_is_at_infinity
        if object_is_at_infinity:
            field = rayfunction_stops.outputs.direction.xy
            pupil = rayfunction_stops.outputs.position.xy
        else:
            field = rayfunction_stops.outputs.position.xy
            pupil = rayfunction_stops.outputs.direction.xy

        if normalized_field:
            min_field = field.min(axis=(axis_field, axis_pupil))
            ptp_field = field.ptp(axis=(axis_field, axis_pupil))
            result.field = ptp_field * (result.field + 1) / 2 + min_field

            if object_is_at_infinity:
                direction = na.Cartesian3dVectorArray(
                    x=result.field.x,
                    y=result.field.y,
                    z=np.sqrt(
                        1 - np.square(result.field.x) - np.square(result.field.y)
                    ),
                )
                result.field = optika.angles(direction)

        if normalized_pupil:
            min_pupil = pupil.min(axis=(axis_field, axis_pupil))
            ptp_pupil = pupil.ptp(axis=(axis_field, axis_pupil))
            result.pupil = ptp_pupil * (result.pupil + 1) / 2 + min_pupil

            if not object_is_at_infinity:
                direction = na.Cartesian3dVectorArray(
                    x=result.pupil.x,
                    y=result.pupil.y,
                    z=np.sqrt(
                        1 - np.square(result.pupil.x) - np.square(result.pupil.y)
                    ),
                )
                result.pupil = optika.angles(direction)

        return result

    def _calc_rayfunction_input(
        self,
        grid_input: optika.vectors.ObjectVectorArray,
        normalized_field: bool = True,
        normalized_pupil: bool = True,
    ) -> optika.rays.RayFunctionArray:

        grid_input = self._denormalize_grid(
            grid=grid_input,
            normalized_field=normalized_field,
            normalized_pupil=normalized_pupil,
        )

        if self.object_is_at_infinity:
            position = grid_input.pupil
            direction = optika.direction(grid_input.field)
        else:
            position = grid_input.field
            direction = optika.direction(grid_input.pupil)

        rays = optika.rays.RayVectorArray(
            wavelength=grid_input.wavelength,
            position=na.Cartesian3dVectorArray(
                x=position.x,
                y=position.y,
                z=0 * position.x.unit,
            ),
            direction=direction,
        )

        obj = self.object
        if obj is not None:
            if obj.transformation is not None:
                rays = obj.transformation(rays)

        result = optika.rays.RayFunctionArray(inputs=grid_input, outputs=rays)

        return result

    @functools.cached_property
    def _rayfunction_input(self) -> optika.rays.RayFunctionArray:
        return self._calc_rayfunction_input(
            grid_input=self.grid_input,
        )

    def raytrace(
        self,
        intensity: None | float | u.Quantity | na.AbstractScalar = None,
        wavelength: None | u.Quantity | na.AbstractScalar = None,
        field: None | na.AbstractCartesian2dVectorArray = None,
        pupil: None | na.AbstractCartesian2dVectorArray = None,
        axis: None | str = None,
        normalized_field: bool = True,
        normalized_pupil: bool = True,
    ) -> optika.rays.RayFunctionArray:
        """
        Given the wavelength, field position, and pupil position of some input
        rays, trace those rays through the system and return the result in
        global coordinates, including all intermediate rays.

        Parameters
        ----------
        intensity
            The energy density of the input rays.
        wavelength
            The wavelengths of the input rays.
            If :obj:`None` (the default), ``self.grid_input.wavelength``
            will be used.
        field
            The field positions of the input rays, in either normalized or physical units.
            If :obj:`None` (the default), ``self.grid_input.field``
            will be used.
        pupil
            The pupil positions of the input rays, in either normalized or physical units.
            If :obj:`None` (the default), ``self.grid_input.pupil``
            will be used.
        axis
            The axis along which the rays are accumulated.
            If :obj:`None` (the default), :attr:`axis_surface` will be used.
        normalized_field
            A boolean flag indicating whether the `field` parameter is given
            in normalized or physical units.
        normalized_pupil
            A boolean flag indicating whether the `pupil` parameter is given
            in normalized or physical units.

        See Also
        --------
        rayfunction : Similar to `raytrace` except it only returns the rays at
            the last surface in local coordinates.
        """

        if axis is None:
            axis = self.axis_surface

        if (wavelength is None) and (field is None) and (pupil is None):
            result = self._rayfunction_input.copy_shallow()
        else:
            grid_input = self.grid_input.copy_shallow()
            if wavelength is not None:
                grid_input.wavelength = wavelength
            if field is not None:
                grid_input.field = field
            if pupil is not None:
                grid_input.pupil = pupil
            result = self._calc_rayfunction_input(
                grid_input,
                normalized_field=normalized_field,
                normalized_pupil=normalized_pupil,
            )

        rays = result.outputs
        if intensity is not None:
            rays.intensity = intensity
        if self.transformation is not None:
            rays = self.transformation.inverse(rays)

        surfaces = self.surfaces_all

        result.outputs = optika.propagators.accumulate_rays(
            propagators=surfaces,
            rays=rays,
            axis=axis,
        )
        return result

    def rayfunction(
        self,
        intensity: None | float | u.Quantity | na.AbstractScalar = None,
        wavelength: None | u.Quantity | na.AbstractScalar = None,
        field: None | na.AbstractCartesian2dVectorArray = None,
        pupil: None | na.AbstractCartesian2dVectorArray = None,
        normalized_field: bool = True,
        normalized_pupil: bool = True,
    ) -> optika.rays.RayFunctionArray:
        """
        Given the wavelength, field position, and pupil position of some input
        rays, trace those rays through the system and return the resulting
        rays in local coordinates at the last surface.

        Parameters
        ----------
        intensity
            The energy density of the input rays.
        wavelength
            The wavelengths of the input rays.
            If :obj:`None` (the default), ``self.grid_input.wavelength``
            will be used.
        field
            The field positions of the input rays, in either normalized or physical units.
            If :obj:`None` (the default), ``self.grid_input.field``
            will be used.
        pupil
            The pupil positions of the input rays, in either normalized or physical units.
            If :obj:`None` (the default), ``self.grid_input.pupil``
            will be used.
        normalized_field
            A boolean flag indicating whether the `field` parameter is given
            in normalized or physical units.
        normalized_pupil
            A boolean flag indicating whether the `pupil` parameter is given
            in normalized or physical units.

        See Also
        --------
        raytrace : Similar to `rayfunction` except it computes all the
            intermediate rays, and it returns results in global coordinates.
        """

        axis = "_dummy"
        raytrace = self.raytrace(
            intensity=intensity,
            wavelength=wavelength,
            field=field,
            pupil=pupil,
            axis=axis,
            normalized_field=normalized_field,
            normalized_pupil=normalized_pupil,
        )
        rayfunction = raytrace[{axis: ~0}]
        rays = rayfunction.outputs

        if self.sensor.transformation is not None:
            rays = self.sensor.transformation.inverse(rays)

        rayfunction.outputs = rays

        return rayfunction

    @functools.cached_property
    def rayfunction_default(self) -> optika.rays.RayFunctionArray:
        """
        Computes the rays in local coordinates at the last surface in the system
        as a function of input wavelength and position using :attr:`grid_input`.

        This property is cached to increase performance.
        If :attr:`grid_input` is updated, the cache must be cleared with
        ``del system.rayfunction_default`` before calling property.
        """
        return self.rayfunction()

    def _rayfunction_from_vertices(
        self,
        radiance: na.AbstractScalar,
        wavelength: na.AbstractScalar,
        field: na.AbstractCartesian2dVectorArray,
        pupil: na.AbstractCartesian2dVectorArray,
        axis_wavelength: str,
        axis_field: tuple[str, str],
        axis_pupil: tuple[str, str],
        normalized_field: bool = True,
        normalized_pupil: bool = True,
    ) -> optika.rays.RayFunctionArray:
        """
        Similar to :meth:`rayfunction` except that the wavelength, field, and
        pupil positions are specified on cell vertices instead of cell centers.

        This function uses the vertices to compute both the area and centroid of
        each cell before calling :meth:`rayfunction` and returning the result.

        Parameters
        ----------
        radiance
            The <spectral radiance https://en.wikipedia.org/wiki/Spectral_radiance>_
            of each cell.
            This will be converted into an energy flux before being passed into
            :meth:`rayfunction`.
        wavelength
            The vertices of the wavelength grid.
        field
            The vertices of the field grid (in either normalized or physical units).
        pupil
            The vertices of the pupil grid (in either normalized or physical units).
        axis_wavelength
            The logical axis corresponding to changing wavelength.
            This axis should only be present in the `wavelength` and `pupil`
            parameters.
        axis_field
            The two logical axes corresponding to changing field positions.
            These axes should only be present in the `field` and `pupil`
            parameters.
        axis_pupil
            The two logical axes corresponding to changing pupil positions.
            These axes should only be present in the `pupil` parameter.
        normalized_field
            A boolean flag indicating if the `field` parameter is in normalized
            units.
        normalized_pupil
            A boolean flag indicating if the `pupil` parameter is in normalized
            units.
        """

        grid = optika.vectors.ObjectVectorArray(
            wavelength=wavelength,
            field=field,
            pupil=pupil,
        )

        grid = self._denormalize_grid(
            grid=grid,
            normalized_field=normalized_field,
            normalized_pupil=normalized_pupil,
        )

        shape = na.broadcast_shapes(self.shape, grid.shape)
        grid_centered = grid.broadcast_to(shape).cell_centers(
            axis=(axis_wavelength,) + axis_field + axis_pupil,
            random=True,
        )

        area = grid.cell_area(
            axis_wavelength=axis_wavelength,
            axis_field=axis_field,
            axis_pupil=axis_pupil,
        )

        flux = radiance * area

        return self.rayfunction(
            intensity=flux,
            wavelength=grid_centered.wavelength,
            field=grid_centered.field,
            pupil=grid_centered.pupil,
            normalized_field=False,
            normalized_pupil=False,
        )

    def image(
        self,
        scene: na.FunctionArray[na.SpectralPositionalVectorArray, na.AbstractScalar],
        wavelength: None | na.AbstractScalar = None,
        pupil: None | na.AbstractCartesian2dVectorArray = None,
        axis_wavelength: None | str = None,
        axis_field: None | tuple[str, str] = None,
        axis_pupil: None | tuple[str, str] = None,
        **kwargs,
    ) -> na.FunctionArray[na.SpectralPositionalVectorArray, na.AbstractScalar]:
        """
        Forward model of the optical system.
        Maps the given spectral radiance of a scene to detector counts.

        Parameters
        ----------
        scene
            The spectral radiance of the scene as a function of wavelength
            and field position.
            The inputs must be cell vertices.
        wavelength
            The vertices of the final wavelength grid on the image plane.
            If :obj:`None` (the default), ``scene.inputs.wavelength`` is used.
        pupil
            The vertices of the pupil grid in either normalized or physical
            coordinates.
            If :obj:`None`, ``self.grid_input.pupil`` is used.
        axis_wavelength
            The logical axis corresponding to changing wavelength coordinate.
            If :obj:`None`,
            ``set(scene.inputs.wavelength.shape) - set(self.shape)``,
            should have only one element.
        axis_field
            The two logical axes corresponding to changing field coordinate.
            If :obj:`None`,
            ``set(scene.inputs.position.shape) - set(self.shape) - {axis_wavelength}``,
            should have exactly two elements.
        axis_pupil
            The two logical axes corresponding to changing pupil coordinate.
            If :obj:`None`,
            ``set(pupil.shape) - set(self.shape) - {axis_wavelength,} - set(axis_field)``,
            should have exactly two elements.
        kwargs
            Additional keyword arguments used by subclass implementations
            of this method.
        """

        shape = self.shape

        scene = scene.explicit

        wavelength_scene = scene.inputs.wavelength
        field = scene.inputs.position

        if wavelength is None:
            wavelength = wavelength_scene

        if pupil is None:
            pupil = self.grid_input.pupil

        unit_field = na.unit_normalized(field)
        unit_pupil = na.unit_normalized(pupil)

        normalized_field = unit_field.is_equivalent(u.dimensionless_unscaled)
        normalized_pupil = unit_pupil.is_equivalent(u.dimensionless_unscaled)

        if axis_wavelength is None:
            axis_wavelength = set(wavelength_scene.shape) - set(shape)
            axis_wavelength = tuple(axis_wavelength)
            if len(axis_wavelength) != 1:  # pragma: nocover
                raise ValueError(
                    "if `axis_wavelength` is `None`, "
                    f"the wavelength axis must be unambiguous, "
                    f"got {axis_wavelength} as possibilities."
                )
            (axis_wavelength,) = axis_wavelength

        if axis_field is None:
            axis_field = set(field.shape) - set(shape)
            axis_field = tuple(axis_field - {axis_wavelength})
            if len(axis_field) != 2:  # pragma: nocover
                raise ValueError(
                    "if `axis_field` is `None`, "
                    "the two field axes must be unambiguous, "
                    f"got {axis_field} as possibilities."
                )

        if axis_pupil is None:
            axis_pupil = set(pupil.shape) - set(shape)
            axis_pupil = tuple(axis_pupil - {axis_wavelength} - set(axis_field))
            if len(axis_pupil) != 2:  # pragma: nocover
                raise ValueError(
                    "if `axis_pupil` is `None`, "
                    "the two pupil axes must be unambiguous, "
                    f"got {axis_pupil} as possibilities."
                )

        rayfunction = self._rayfunction_from_vertices(
            radiance=scene.outputs,
            wavelength=wavelength_scene,
            field=field,
            pupil=pupil,
            axis_wavelength=axis_wavelength,
            axis_field=axis_field,
            axis_pupil=axis_pupil,
            normalized_field=normalized_field,
            normalized_pupil=normalized_pupil,
        )

        return self.sensor.readout(
            rays=rayfunction.outputs,
            wavelength=wavelength,
            axis=(axis_wavelength,) + axis_field + axis_pupil,
        )

    def plot(
        self,
        ax: None | matplotlib.axes.Axes | na.ScalarArray[npt.NDArray] = None,
        transformation: None | na.transformations.AbstractTransformation = None,
        components: None | tuple[str, ...] = None,
        plot_rays: bool = True,
        plot_rays_vignetted: bool = False,
        kwargs_rays: None | dict[str, Any] = None,
        **kwargs,
    ) -> na.AbstractScalar | dict[str, na.AbstractScalar]:
        """
        Plot the surfaces of the system and the default raytrace.

        Parameters
        ----------
        ax
            The matplotlib axes on which to plot the system
        transformation
            Any additional transformation to apply to the system before plotting.
        components
            The vector components to plot if `ax` is 2-dimensional.
        plot_rays
            Boolean flag indicating whether to plot the rays.
        plot_rays_vignetted
            Boolean flag indicating whether to plot the vignetted rays.
        kwargs_rays
            Any additional keyword arguments to use when plotting the rays.
        kwargs
            Any additional keyword arguments to use when plotting the surfaces.
        """

        surfaces = self.surfaces_all
        transformation_self = self.transformation
        kwargs_plot = self.kwargs_plot

        if transformation is not None:
            if transformation_self is not None:
                transformation = transformation @ transformation_self
        else:
            transformation = transformation_self

        if kwargs_plot is not None:
            kwargs = kwargs | kwargs_plot

        result = dict()

        result["surfaces"] = []
        for surface in surfaces:
            result["surfaces"].append(
                surface.plot(
                    ax=ax,
                    transformation=transformation,
                    components=components,
                    **kwargs,
                )
            )

        if plot_rays:
            raytrace = self.raytrace()

            if kwargs_rays is None:
                kwargs_rays = dict()

            where = True
            if not plot_rays_vignetted:
                where = raytrace.outputs.unvignetted[{self.axis_surface: ~0}]

            result["rays"] = na.plt.plot(
                raytrace.outputs.position,
                ax=ax,
                axis=self.axis_surface,
                where=where,
                transformation=transformation,
                components=components,
                **kwargs_rays,
            )

        return result


@dataclasses.dataclass(eq=False, repr=False)
class SequentialSystem(
    AbstractSequentialSystem,
):
    """
    A sequential optical system is a composition of a sequence of
    :class:`optika.surfaces.AbstractSurface` instances and a default
    grid to sample them with.

    Examples
    --------

    Here is an example of a simple Newtonian telescope, with a parabolic
    primary mirror, a 45 degree fold mirror, and a detector.

    .. jupyter-execute::

        import dataclasses
        import matplotlib.pyplot as plt
        import astropy.units as u
        import astropy.visualization
        import named_arrays as na
        import optika

        # store the coordinates of the primary mirror, fold mirror,
        # and sensor, so we can determine the focal length of the
        # primary mirror.
        primary_mirror_z = 200 * u.mm
        fold_mirror_z = 50 * u.mm
        sensor_x = 50 * u.mm

        # Define the front aperture surface.
        front = optika.surfaces.Surface(
            name="front",
        )

        # Define the parabolic primary mirror.
        primary_mirror = optika.surfaces.Surface(
            name="mirror",
            sag=optika.sags.ParabolicSag(
                focal_length=-(primary_mirror_z - fold_mirror_z + sensor_x),
            ),
            aperture=optika.apertures.RectangularAperture(40 * u.mm),
            material=optika.materials.Mirror(),
            is_pupil_stop=True,
            transformation=na.transformations.Cartesian3dTranslation(
                z=primary_mirror_z,
            ),
        )

        # Define the flat fold mirror which directs light
        # to the detector surface.
        fold_mirror = optika.surfaces.Surface(
            name="fold_mirror",
            aperture=optika.apertures.RectangularAperture(25 * u.mm),
            material=optika.materials.Mirror(),
            transformation=na.transformations.TransformationList([
                na.transformations.Cartesian3dRotationY((90 + 45) * u.deg),
                na.transformations.Cartesian3dTranslation(
                    z=fold_mirror_z,
                ),
            ]),
        )

        # Define the central obscuration, the back face
        # of the fold mirror which blocks some of the
        # incoming light.
        obscuration = optika.surfaces.Surface(
            name="obscuration",
            aperture=dataclasses.replace(fold_mirror.aperture, inverted=True),
            transformation=fold_mirror.transformation,
        )

        # define the imaging sensor surface
        sensor = optika.sensors.ImagingSensor(
            name="sensor",
            width_pixel=20 * u.um,
            axis_pixel=na.Cartesian2dVectorArray("detector_x", "detector_y"),
            num_pixel=na.Cartesian2dVectorArray(128, 128),
            timedelta_exposure=1 * u.s,
            transformation=na.transformations.TransformationList([
                na.transformations.Cartesian3dRotationY(-90 * u.deg),
                na.transformations.Cartesian3dTranslation(
                    x=-sensor_x,
                    z=fold_mirror_z,
                ),
            ]),
            is_field_stop=True,
        )

        # Define the grid of normalized field coordinates,
        field = na.Cartesian2dVectorLinearSpace(
            start=-1,
            stop=1,
            axis=na.Cartesian2dVectorArray("field_x", "field_y"),
            num=5,
            centers=True,
        )

        # Define the grid of normalized pupil coordinates
        # in a similar fashion to the normalized field
        # coordinates
        pupil = na.Cartesian2dVectorLinearSpace(
            start=-1,
            stop=1,
            axis=na.Cartesian2dVectorArray("pupil_x", "pupil_y"),
            num=5,
            centers=True,
        )

        # define the optical system using the surfaces
        # and the normalized field/pupil coordinates
        system = optika.systems.SequentialSystem(
            surfaces=[
                front,
                obscuration,
                primary_mirror,
                fold_mirror,
            ],
            sensor=sensor,
            grid_input=optika.vectors.ObjectVectorArray(
                wavelength=500 * u.nm,
                field=field,
                pupil=pupil,
            ),
        )

        # plot the system
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(constrained_layout=True)
            ax.set_aspect("equal")
            system.plot(
                ax=ax,
                components=("z", "x"),
                kwargs_rays=dict(
                    color="tab:blue",
                ),
                color="black",
                zorder=10,
            )

    |

    Using this model, we can simulate an image of an airforce target

    .. jupyter-execute::

        # Define the number of points to sample
        num_field = 2 * system.sensor.num_pixel

        # Define the scene as an airforce target.
        # Note how the coordinates (inputs) are defined on
        # cell vertices and the values (outputs) are
        # defined on cell centers.
        scene = na.FunctionArray(
            inputs=na.SpectralPositionalVectorArray(
                wavelength=na.linspace(530, 531, axis="wavelength", num=2) * u.nm,
                position=na.Cartesian2dVectorLinearSpace(
                    start=system.field_min,
                    stop=system.field_max,
                    axis=na.Cartesian2dVectorArray("field_x", "field_y"),
                    num=num_field + 1,
                ),
            ),
            outputs=optika.targets.airforce(
                axis_x="field_x",
                axis_y="field_y",
                num_x=num_field.x,
                num_y=num_field.y,
            ) * 100 * u.photon / u.s / u.m ** 2 / u.arcsec ** 2 / u.nm,
        )

        # Simulate an image of the scene using the optical system
        image = system.image(scene)

        # Plot the original scene and the simulated image
        with astropy.visualization.quantity_support():
            fig, ax = plt.subplots(
                ncols=2,
                figsize=(8, 5),
                constrained_layout=True,
            )
            mappable_scene = na.plt.pcolormesh(
                scene.inputs.position,
                C=scene.outputs.value,
                ax=ax[0],
            )
            mappable_image = na.plt.pcolormesh(
                image.inputs.position,
                C=image.outputs.value.sum("wavelength"),
                ax=ax[1],
            )
            cbar_0 = fig.colorbar(
                mappable=mappable_scene.ndarray.item(),
                ax=ax[0],
                location="top",
            )
            cbar_0.set_label(f"radiance ({scene.outputs.unit:latex_inline})")
            cbar_1 = fig.colorbar(
                mappable=mappable_image.ndarray.item(),
                ax=ax[1],
                location="top",
            )
            cbar_1.set_label(f"measured charge ({image.outputs.unit:latex_inline})")
            ax[0].set_aspect("equal")
            ax[1].set_aspect("equal")

    The result is flipped vertically and horizontally due to the layout
    of the optical system.
    The noise on the image is from the stratified random sampling used to
    generate the grid of rays traced through the system, there is no
    additional noise sources, such as photon shot noise in this simulation.
    """

    surfaces: Sequence[optika.surfaces.AbstractSurface] = dataclasses.MISSING
    """
    a sequence of surfaces representing this optical system.

    At least one of these surfaces needs to be marked as the pupil surface,
    and if the object surface is not marked as the field stop, one of these
    surfaces needs to be marked as the field stop.
    """

    object: optika.surfaces.AbstractSurface = None
    """
    The external object being imaged or illuminated by this system.

    If :obj:`None`, the object is assumed to be an empty surface at the origin.
    """

    sensor: optika.sensors.AbstractImagingSensor = None
    """
    The imaging sensor that measures the light captured by this system.

    This is the last surface in the optical system.
    """

    axis_surface: str = "surface"
    """
    The name of the logical axis representing the sequence of surfaces.
    """

    grid_input: optika.vectors.ObjectVectorArray = None
    """
    The input grid to sample with rays.

    This grid is simultaneously projected onto both the field stop and the
    pupil stop.

    Positions on the stop can be specified in either absolute or normalized
    coordinates. Using normalized coordinates allows for injecting different
    grid types (cylindrical, stratified random, etc.) without specifying the
    scale of the stop surface.

    If positions are specified in absolute units, they are measured in the
    coordinate system of the corresponding stop surface.
    """

    transformation: None | na.transformations.AbstractTransformation = None
    """
    A optional coordinate transformation to apply to the entire optical system.
    """

    kwargs_plot: None | dict[str, Any] = None
    """
    Additional keyword arguments used by default in :meth:`plot`.
    """

    def __post_init__(self):
        if self.object is None:
            self.object = optika.surfaces.Surface()

    @property
    def shape(self) -> dict[str, int]:
        return na.broadcast_shapes(
            *[optika.shape(surface) for surface in self.surfaces],
            optika.shape(self.object),
            optika.shape(self.sensor),
            optika.shape(self.transformation),
        )



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika


@pytest.mark.parametrize(
    argnames="angles",
    argvalues=[
        na.Cartesian2dVectorArray(1, 2) * u.deg,
        na.Cartesian2dVectorLinearSpace(
            start=0,
            stop=1,
            axis=na.Cartesian2dVectorArray("x", "y"),
            num=5,
        ),
    ],
)
def test_direction(angles: na.AbstractCartesian2dVectorArray):
    result = optika.direction(angles)

    rotation_x = na.Cartesian3dYRotationMatrixArray(-angles.x)
    rotation_y = na.Cartesian3dXRotationMatrixArray(+angles.y)
    result_expected = rotation_x @ rotation_y @ na.Cartesian3dVectorArray(z=1)
    # result_expected = rotation_y @ rotation_x @ na.Cartesian3dVectorArray(z=1)

    print(f"{result=}")
    print(f"{result_expected=}")

    assert isinstance(result, na.AbstractCartesian3dVectorArray)
    assert np.allclose(result, result_expected)


@pytest.mark.parametrize(
    argnames="direction",
    argvalues=[
        na.Cartesian3dVectorArray(1, 2, 5).normalized,
    ],
)
def test_angles(direction: na.AbstractCartesian3dVectorArray):
    result = optika.angles(direction)

    print(f"{direction=}")
    print(f"{optika.direction(result)=}")

    assert isinstance(result, na.AbstractCartesian2dVectorArray)
    assert np.allclose(direction, optika.direction(result))



import pytest
import abc
import dataclasses
import numpy as np
import matplotlib.axes
import matplotlib.pyplot as plt
import astropy.units as u
import named_arrays as na
import optika


transformation_parameterization = [
    None,
    na.transformations.TransformationList(
        [
            na.transformations.Cartesian3dTranslation(x=5 * u.mm),
            na.transformations.Cartesian3dRotationZ(53 * u.deg),
            na.transformations.Cartesian3dTranslation(x=6 * u.mm),
        ]
    ),
    na.transformations.Cartesian3dRotationZ(
        na.ScalarLinearSpace(0 * u.deg, 90 * u.deg, axis="transformation", num=3)
    ),
    na.transformations.Cartesian3dRotationZ(
        na.NormalUncertainScalarArray(53 * u.deg, width=5 * u.deg)
    ),
]

kwargs_plot_parameterization = [
    dict(),
    dict(color="red"),
]


class AbstractTestShaped(
    abc.ABC,
):
    def test_shape(self, a: optika.mixins.Shaped):
        result = a.shape
        assert isinstance(result, dict)
        for k in result:
            assert isinstance(k, str)
            assert isinstance(result[k], int)


class AbstractTestPrintable(
    abc.ABC,
):
    @pytest.mark.parametrize(
        argnames="prefix",
        argvalues=[None, "   "],
    )
    def test_to_string(
        self,
        a: optika.mixins.Printable,
        prefix: None | str,
    ):
        result = a.to_string(prefix=prefix)
        assert isinstance(result, str)


class AbstractTestPlottable(abc.ABC):
    @pytest.mark.parametrize(
        argnames="ax",
        argvalues=[
            None,
            plt.subplots()[1],
        ],
    )
    @pytest.mark.parametrize(
        argnames="transformation",
        argvalues=transformation_parameterization[:2],
    )
    class TestPlot(abc.ABC):
        def test_plot(
            self,
            a: optika.mixins.Plottable,
            ax: None | matplotlib.axes.Axes | na.ScalarArray,
            transformation: None | na.transformations.AbstractTransformation,
        ):
            result = a.plot(
                ax=ax,
                transformation=transformation,
            )

            if ax is None or ax is np._NoValue:
                ax_normalized = plt.gca()
            else:
                ax_normalized = ax
            ax_normalized = na.as_named_array(ax_normalized)

            for index in ax_normalized.ndindex():
                assert ax_normalized[index].ndarray.has_data()

            assert isinstance(result, (na.AbstractScalar, dict))


class AbstractTestTransformable:
    def test_transformation(self, a: optika.mixins.Transformable):
        t = a.transformation
        if t is not None:
            assert isinstance(t, na.transformations.AbstractTransformation)


class AbstractTestTranslatable(
    AbstractTestTransformable,
):
    def test_translation(
        self,
        a: optika.mixins.Translatable,
    ):
        result = a.translation
        assert na.unit_normalized(result).is_equivalent(u.mm)


@dataclasses.dataclass(eq=False, repr=False)
class Translatable(
    optika.mixins.Translatable,
):
    translation: u.Quantity | na.AbstractCartesian3dVectorArray = 0 * u.mm


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        Translatable(),
        Translatable(na.Cartesian3dVectorArray(1, 2, 3) * u.mm),
    ],
)
class TestTranslatable(
    AbstractTestTranslatable,
):
    pass


class AbstractTestPitchable(
    AbstractTestTransformable,
):
    def test_pitch(
        self,
        a: optika.mixins.Pitchable,
    ):
        result = a.pitch
        assert isinstance(na.as_named_array(result), na.AbstractScalar)
        assert na.unit_normalized(result).is_equivalent(u.deg)


@dataclasses.dataclass(eq=False, repr=False)
class Pitchable(
    optika.mixins.Pitchable,
):
    pitch: u.Quantity | na.AbstractScalar = 0 * u.deg


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        Pitchable(),
        Pitchable(10 * u.deg),
    ],
)
class TestPitchable(
    AbstractTestPitchable,
):
    pass


class AbstractTestYawable(
    AbstractTestTransformable,
):
    def test_yaw(
        self,
        a: optika.mixins.Yawable,
    ):
        result = a.yaw
        assert isinstance(na.as_named_array(result), na.AbstractScalar)
        assert na.unit_normalized(result).is_equivalent(u.deg)


@dataclasses.dataclass(eq=False, repr=False)
class Yawable(
    optika.mixins.Yawable,
):
    yaw: u.Quantity | na.AbstractScalar = 0 * u.deg


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        Yawable(),
        Yawable(10 * u.deg),
    ],
)
class TestYawable(
    AbstractTestYawable,
):
    pass


class AbstractTestRollable(
    AbstractTestTransformable,
):
    def test_roll(
        self,
        a: optika.mixins.Rollable,
    ):
        result = a.roll
        assert isinstance(na.as_named_array(result), na.AbstractScalar)
        assert na.unit_normalized(result).is_equivalent(u.deg)


@dataclasses.dataclass(eq=False, repr=False)
class Rollable(
    optika.mixins.Rollable,
):
    roll: u.Quantity | na.AbstractScalar = 0 * u.deg


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        Rollable(),
        Rollable(10 * u.deg),
    ],
)
class TestRollable(
    AbstractTestRollable,
):
    pass



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika._tests.test_surfaces import AbstractTestAbstractSurface


class AbstractTestAbstractImagingSensor(
    AbstractTestAbstractSurface,
):
    def test_num_pixel(self, a: optika.sensors.AbstractImagingSensor):
        result = a.num_pixel
        assert isinstance(result, na.AbstractCartesian2dVectorArray)
        assert np.issubdtype(na.get_dtype(result.x), int)
        assert np.issubdtype(na.get_dtype(result.y), int)

    def test_timedelta_exposure(self, a: optika.sensors.AbstractImagingSensor):
        result = a.timedelta_exposure
        assert result >= 0 * u.s

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=[
            optika.rays.RayVectorArray(
                intensity=100 * u.photon / u.s,
                wavelength=500 * u.nm,
                position=na.Cartesian3dVectorArray() * u.mm,
            ),
            optika.rays.RayVectorArray(
                intensity=na.random.poisson(100, shape_random=dict(t=11)) * u.erg / u.s,
                wavelength=500 * u.nm,
                position=na.Cartesian3dVectorArray(
                    x=na.random.uniform(-1, 1, shape_random=dict(t=11)) * u.mm,
                    y=na.random.uniform(-1, 1, shape_random=dict(t=11)) * u.mm,
                    z=0 * u.mm,
                ),
            ),
        ],
    )
    @pytest.mark.parametrize(
        argnames="wavelength",
        argvalues=[
            na.linspace(500, 600, axis="wavelength", num=11) * u.nm,
        ],
    )
    def test_readout(
        self,
        a: optika.sensors.AbstractImagingSensor,
        rays: optika.rays.RayVectorArray,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.readout(rays, wavelength)
        assert isinstance(result, na.FunctionArray)
        assert isinstance(result.inputs, na.SpectralPositionalVectorArray)
        assert isinstance(result.outputs, na.AbstractScalar)
        assert result.outputs.unit.is_equivalent(u.electron)
        assert a.axis_pixel.x in result.outputs.shape
        assert a.axis_pixel.y in result.outputs.shape


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.sensors.ImagingSensor(
            name="test sensor",
            width_pixel=15 * u.um,
            axis_pixel=na.Cartesian2dVectorArray("detector_x", "detector_y"),
            num_pixel=na.Cartesian2dVectorArray(2048, 1024),
            transformation=na.transformations.Cartesian3dTranslation(x=1 * u.mm),
        ),
    ],
)
class TestImagingSensor(
    AbstractTestAbstractImagingSensor,
):
    pass



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika


@pytest.mark.parametrize(
    argnames="absorption",
    argvalues=[
        1 / u.um,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_substrate",
    argvalues=[
        15 * u.um,
    ],
)
@pytest.mark.parametrize(
    argnames="thickness_depletion",
    argvalues=[
        5 * u.um,
    ],
)
def test_charge_diffusion(
    absorption: u.Quantity | na.AbstractScalar,
    thickness_substrate: u.Quantity | na.AbstractScalar,
    thickness_depletion: u.Quantity | na.AbstractScalar,
):
    result = optika.sensors.charge_diffusion(
        absorption=absorption,
        thickness_substrate=thickness_substrate,
        thickness_depletion=thickness_depletion,
    )

    assert result > 0 * u.um


@pytest.mark.parametrize(
    argnames="width_diffusion",
    argvalues=[
        10 * u.um,
        na.linspace(1, 10, "width", 5) * u.um,
    ],
)
@pytest.mark.parametrize(
    argnames="width_pixel",
    argvalues=[
        15 * u.um,
    ],
)
def test_mean_charge_capture(
    width_diffusion: u.Quantity | na.AbstractScalar,
    width_pixel: u.Quantity | na.AbstractScalar,
):
    result = optika.sensors.mean_charge_capture(
        width_diffusion=width_diffusion,
        width_pixel=width_pixel,
    )
    assert np.all(result > 0)
    assert np.all(result < 1)



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
from named_arrays._vectors.cartesian.tests import test_vectors_cartesian
import optika

_num_y = test_vectors_cartesian._num_y


vectors_polarization = [
    optika.vectors.PolarizationVectorArray(
        s=1,
        p=0,
    ),
]


def _items() -> list[na.AbstractArray | dict[str, int | slice | na.AbstractArray]]:
    return [
        dict(y=0),
        dict(y=slice(0, 1)),
        dict(y=na.ScalarArrayRange(0, 2, axis="y")),
    ]


class AbstractTestAbstractPolarizationVectorArray(
    test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray
):
    def test_s(self, array: optika.vectors.AbstractPolarizationVectorArray):
        assert isinstance(na.as_named_array(array.s), na.AbstractScalar)

    def test_p(self, array: optika.vectors.AbstractPolarizationVectorArray):
        assert isinstance(na.as_named_array(array.p), na.AbstractScalar)

    def test_average(self, array: optika.vectors.AbstractPolarizationVectorArray):
        assert isinstance(na.as_named_array(array.average), na.AbstractScalar)

    @pytest.mark.xfail(raises=NotImplementedError)
    def test_matrix(self, array: optika.vectors.AbstractPolarizationVectorArray):
        return super().test_matrix(array=array)

    @pytest.mark.parametrize(
        argnames="item",
        argvalues=_items(),
    )
    def test__getitem__(
        self,
        array: optika.vectors.AbstractPolarizationVectorArray,
        item: dict[str, int | slice | na.AbstractArray] | na.AbstractArray,
    ):
        super().test__getitem__(array=array, item=item)

    @pytest.mark.parametrize("array_2", vectors_polarization)
    class TestUfuncBinary(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestUfuncBinary
    ):
        pass

    @pytest.mark.parametrize("array_2", vectors_polarization)
    class TestMatmul(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestMatmul
    ):
        pass

    class TestArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions
    ):
        @pytest.mark.parametrize("array_2", vectors_polarization)
        class TestAsArrayLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestAsArrayLikeFunctions
        ):
            pass

        @pytest.mark.parametrize(
            argnames="where",
            argvalues=[
                np._NoValue,
            ],
        )
        class TestReductionFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestReductionFunctions,
        ):
            pass

        @pytest.mark.parametrize(
            argnames="q",
            argvalues=[
                25 * u.percent,
            ],
        )
        class TestPercentileLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestArrayFunctions.TestPercentileLikeFunctions,
        ):
            pass

    class TestNamedArrayFunctions(
        test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions,
    ):
        @pytest.mark.skip
        class TestPltPlotLikeFunctions(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions.TestPltPlotLikeFunctions,
        ):
            pass

        @pytest.mark.skip
        class TestJacobian(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions.TestJacobian,
        ):
            pass

        @pytest.mark.skip
        class TestOptimizeRoot(
            test_vectors_cartesian.AbstractTestAbstractCartesianVectorArray.TestNamedArrayFunctions.TestOptimizeRoot,
        ):
            pass


@pytest.mark.parametrize("array", vectors_polarization)
class TestPolarizationVectorArray(
    AbstractTestAbstractPolarizationVectorArray,
    test_vectors_cartesian.AbstractTestAbstractExplicitCartesianVectorArray,
):
    @pytest.mark.parametrize(
        argnames="item",
        argvalues=[
            dict(y=0),
            dict(y=slice(None)),
        ],
    )
    @pytest.mark.parametrize(
        argnames="value",
        argvalues=[
            optika.vectors.PolarizationVectorArray(
                s=0,
                p=0,
            ),
        ],
    )
    def test__setitem__(
        self,
        array: optika.vectors.AbstractPolarizationVectorArray,
        item: dict[str, int | slice | na.ScalarArray] | na.ScalarArray,
        value: optika.vectors.AbstractPolarizationVectorArray,
    ):
        super().test__setitem__(array=array, item=item, value=value)



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika

_wavelength = na.linspace(100, 200, axis="wavelength", num=4) * u.AA


@pytest.mark.parametrize("wavelength", [_wavelength])
@pytest.mark.parametrize("direction_left", [1])
@pytest.mark.parametrize("direction_right", [1])
@pytest.mark.parametrize("polarized_s", [False, True])
@pytest.mark.parametrize("n_left", [1])
@pytest.mark.parametrize("n_right", [optika.chemicals.Chemical("Si").n(_wavelength)])
@pytest.mark.parametrize(
    argnames="interface",
    argvalues=[optika.materials.profiles.ErfInterfaceProfile(5 * u.nm)],
)
def test_refraction(
    wavelength: u.Quantity | na.AbstractScalar,
    direction_left: float | na.AbstractScalar,
    direction_right: float | na.AbstractScalar,
    polarized_s: bool | na.AbstractScalar,
    n_left: float | na.AbstractScalar,
    n_right: float | na.AbstractScalar,
    interface: None | optika.materials.profiles.AbstractInterfaceProfile,
):
    result = optika.materials.matrices.refraction(
        wavelength=wavelength,
        direction_left=direction_left,
        direction_right=direction_right,
        polarized_s=polarized_s,
        n_left=n_left,
        n_right=n_right,
        interface=interface,
    )
    assert isinstance(result, na.AbstractCartesian2dMatrixArray)
    assert np.all(result.determinant != 0)


@pytest.mark.parametrize("wavelength", [_wavelength])
@pytest.mark.parametrize("direction", [1])
@pytest.mark.parametrize("thickness", [10 * u.nm])
@pytest.mark.parametrize("n", [optika.chemicals.Chemical("Si").n(_wavelength)])
def test_propagation(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: float | na.AbstractScalar,
    thickness: u.Quantity | na.AbstractScalar,
    n: u.Quantity | na.AbstractScalar,
):
    result = optika.materials.matrices.propagation(
        wavelength=wavelength,
        direction=direction,
        thickness=thickness,
        n=n,
    )
    assert isinstance(result, na.AbstractCartesian2dMatrixArray)
    assert np.all(result.determinant != 0)


@pytest.mark.parametrize("wavelength", [_wavelength])
@pytest.mark.parametrize("direction", [na.Cartesian3dVectorArray(0, 0, 1)])
@pytest.mark.parametrize("polarized_s", [False, True])
@pytest.mark.parametrize("thickness", [10 * u.nm])
@pytest.mark.parametrize("n", [optika.chemicals.Chemical("Si").n(_wavelength)])
@pytest.mark.parametrize("normal", [na.Cartesian3dVectorArray(0, 0, -1)])
def test_transfer(
    wavelength: u.Quantity | na.AbstractScalar,
    direction: na.AbstractCartesian3dVectorArray,
    polarized_s: bool | na.AbstractScalar,
    thickness: u.Quantity | na.AbstractScalar,
    n: float | na.AbstractScalar,
    normal: na.AbstractCartesian3dVectorArray,
):
    result = optika.materials.matrices.transfer(
        wavelength=wavelength,
        direction=direction,
        polarized_s=polarized_s,
        thickness=thickness,
        n=n,
        normal=normal,
    )
    assert isinstance(result, na.AbstractCartesian2dMatrixArray)
    assert np.all(result.determinant != 0)



from __future__ import annotations
from typing import TypeVar
from typing_extensions import Self
import dataclasses
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from . import AbstractPupilVectorArray, PupilVectorArray
from . import AbstractSceneVectorArray, SceneVectorArray

__all__ = [
    "AbstractObjectVectorArray",
    "ObjectVectorArray",
]


WavelengthT = TypeVar("WavelengthT", bound=na.ScalarLike)
FieldT = TypeVar("FieldT", bound=na.AbstractCartesian2dVectorArray)
PupilT = TypeVar("PupilT", bound=na.AbstractCartesian2dVectorArray)


@dataclasses.dataclass(eq=False, repr=False)
class AbstractObjectVectorArray(
    AbstractPupilVectorArray,
    AbstractSceneVectorArray,
):
    @property
    def type_abstract(self) -> type[AbstractObjectVectorArray]:
        return AbstractObjectVectorArray

    @property
    def type_explicit(self) -> type[ObjectVectorArray]:
        return ObjectVectorArray

    @property
    def type_matrix(self) -> type[na.AbstractMatrixArray]:
        raise NotImplementedError

    def cell_area(
        self,
        axis_wavelength: str,
        axis_field: tuple[str, str],
        axis_pupil: tuple[str, str],
    ) -> na.AbstractScalar:
        r"""
        Compute the 5-dimensional area of each grid cell in units of wavelength
        :math:`\times` area :math:`\times` solid angle.

        This method does not work in general.
        It only works for the special case where the wavelength grid
        is not dependent on the field or pupil position,
        and the field grid does not depend on the pupil position.

        Parameters
        ----------
        axis_wavelength
            The logical axis corresponding to changing wavelength.
        axis_field
            The two logical axes corresponding to changing field position.
        axis_pupil
            the two logical axes corresponding to changing pupil position.
        """

        wavelength = self.wavelength
        field = self.field
        pupil = self.pupil

        shape_wavelength = wavelength.shape
        shape_field = field.shape
        shape_pupil = pupil.shape

        if axis_wavelength not in shape_wavelength:
            raise ValueError(  # pragma: nocover
                f"{axis_wavelength=} must be in {shape_wavelength=}",
            )
        if not set(axis_field).issubset(shape_field):
            raise ValueError(  # pragma: nocover
                f"{axis_field=} must be a subset of {shape_field=}",
            )
        if set(axis_field).intersection(shape_wavelength):
            raise ValueError(  # pragma: nocover
                f"{axis_field=} must not intersect {shape_wavelength=}"
            )
        if not set(axis_pupil).issubset(shape_pupil):
            raise ValueError(  # pragma: nocover
                f"{axis_pupil=} must be a subset of {shape_pupil=}",
            )
        if set(axis_pupil).intersection(shape_wavelength | shape_field):
            raise ValueError(  # pragma: nocover
                f"{axis_pupil=} must not intersect {shape_wavelength=} or {shape_field=}"
            )

        area_wavelength = wavelength.volume_cell(axis_wavelength)

        shape_field = na.broadcast_shapes(
            shape_wavelength,
            shape_field,
        )
        field = field.broadcast_to(shape_field)

        shape_pupil = na.broadcast_shapes(
            shape_wavelength,
            shape_field,
            shape_pupil,
        )
        pupil = pupil.broadcast_to(shape_pupil)

        if na.unit_normalized(field).is_equivalent(u.deg):
            area_field = optika.direction(field).solid_angle_cell(axis_field)
        else:
            area_field = field.volume_cell(axis_field)

        if na.unit_normalized(pupil).is_equivalent(u.deg):
            area_pupil = optika.direction(pupil).solid_angle_cell(axis_pupil)
        else:
            area_pupil = pupil.volume_cell(axis_pupil)

        area_field = np.abs(area_field)
        area_pupil = np.abs(area_pupil)

        area_field = area_field.cell_centers(
            axis=axis_wavelength,
        )

        area_pupil = area_pupil.cell_centers(
            axis=(axis_wavelength,) + axis_field,
        )

        return area_wavelength * area_field * area_pupil


@dataclasses.dataclass(eq=False, repr=False)
class ObjectVectorArray(
    PupilVectorArray,
    SceneVectorArray,
    AbstractObjectVectorArray,
):
    @classmethod
    def from_scalar(
        cls,
        scalar: na.AbstractScalar,
        like: None | Self = None,
    ) -> Self:
        if like is not None:
            return type(like)(
                wavelength=scalar,
                field=scalar,
                pupil=scalar,
            )
        else:
            return cls(
                wavelength=scalar,
                field=scalar,
                pupil=scalar,
            )



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
import optika._tests.test_mixins
import optika.rays._tests.test_ray_vectors

_wavelength = na.linspace(100, 300, axis="wavelength", num=11) * u.AA


class AbstractTestAbstractMaterial(
    optika._tests.test_mixins.AbstractTestTransformable,
    optika._tests.test_mixins.AbstractTestShaped,
):
    def test_is_mirror(self, a: optika.materials.AbstractMaterial):
        assert isinstance(a.is_mirror, bool)

    @pytest.mark.parametrize("rays", optika.rays._tests.test_ray_vectors.rays)
    class TestRayDependentMethods:
        def test_index_refraction(
            self,
            a: optika.materials.AbstractMaterial,
            rays: optika.rays.AbstractRayVectorArray,
        ):
            result = a.index_refraction(rays)
            assert isinstance(na.as_named_array(result), na.AbstractScalar)
            assert na.unit_normalized(result).is_equivalent(u.dimensionless_unscaled)

        def test_attenuation(
            self,
            a: optika.materials.AbstractMaterial,
            rays: optika.rays.AbstractRayVectorArray,
        ):
            result = a.attenuation(rays)
            assert isinstance(na.as_named_array(result), na.AbstractScalar)
            assert na.unit_normalized(result).is_equivalent(1 / u.mm)

        @pytest.mark.parametrize(
            argnames="normal",
            argvalues=[
                na.Cartesian3dVectorArray(0, 0, -1),
            ],
        )
        def test_efficiency(
            self,
            a: optika.materials.AbstractMaterial,
            rays: optika.rays.AbstractRayVectorArray,
            normal: na.AbstractCartesian3dVectorArray,
        ):
            result = a.efficiency(rays, normal)
            assert isinstance(na.as_named_array(result), na.AbstractScalar)
            assert na.unit_normalized(result).is_equivalent(u.dimensionless_unscaled)
            assert np.all(result >= 0)
            assert np.all(result <= 1)


@pytest.mark.parametrize("a", [optika.materials.Vacuum()])
class TestVacuum(
    AbstractTestAbstractMaterial,
):
    pass


class AbstractTestAbstractMirror(
    AbstractTestAbstractMaterial,
):
    def test_substrate(self, a: optika.materials.AbstractMirror):
        result = a.substrate
        if result is not None:
            assert isinstance(result, optika.materials.Layer)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.Mirror(),
        optika.materials.Mirror(
            substrate=optika.materials.Layer(
                chemical="SiO2",
                thickness=10 * u.mm,
            ),
        ),
    ],
)
class TestMirror(
    AbstractTestAbstractMirror,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.MeasuredMirror(
            efficiency_measured=na.FunctionArray(
                inputs=na.SpectralDirectionalVectorArray(
                    wavelength=_wavelength,
                    direction=na.Cartesian3dVectorArray(0, 0, 1),
                ),
                outputs=np.exp(-np.square(_wavelength / (10 * u.AA)) / 2),
            ),
        )
    ],
)
class TestMeasuredMirror(
    AbstractTestAbstractMirror,
):
    pass



import pytest
import astropy.units as u
import optika
from optika._tests import test_mixins


class AbstractTestAbstractMesh(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestShaped,
):
    def test_chemical(self, a: optika.materials.meshes.AbstractMesh):
        result = a.chemical
        assert isinstance(result, (str, optika.chemicals.AbstractChemical))

    def test_efficiency(self, a: optika.materials.meshes.AbstractMesh):
        result = a.efficiency
        assert result > 0

    def test_pitch(self, a: optika.materials.meshes.AbstractMesh):
        result = a.pitch
        assert result > (0 / u.mm)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.materials.meshes.Mesh(
            chemical="Ni",
            efficiency=0.8,
            pitch=70 / u.mm,
        )
    ],
)
class TestMesh(
    AbstractTestAbstractMesh,
):
    pass



import pytest
import abc
import pathlib
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from optika._tests import test_mixins


_wavelength = [
    500 * u.nm,
    na.geomspace(10, 10000, axis="wavelength", num=101) * u.AA,
    na.NormalUncertainScalarArray(500 * u.nm, width=10 * u.nm),
]


class AbstractTestAbstractChemical(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestShaped,
    abc.ABC,
):
    def test_formula(self, a: optika.chemicals.AbstractChemical):
        result = a.formula
        assert isinstance(result, (str, na.AbstractScalar))

    def test_formula_latex(self, a: optika.chemicals.AbstractChemical):
        result = a.formula_latex
        assert isinstance(result, (str, na.AbstractScalar))

    def test_is_amorphous(self, a: optika.chemicals.AbstractChemical):
        result = a.is_amorphous
        assert np.issubdtype(na.get_dtype(result), bool)

    def test_table(self, a: optika.chemicals.AbstractChemical):
        result = a.table
        if result is not None:
            assert isinstance(result, str)

    def test_file_nk(self, a: optika.chemicals.AbstractChemical):
        result = a.file_nk
        for index in result.ndindex():
            assert isinstance(result[index].ndarray, pathlib.Path)
            assert result[index].ndarray.exists()

    @pytest.mark.parametrize("wavelength", _wavelength)
    def test_index_refraction(
        self,
        a: optika.chemicals.AbstractChemical,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.index_refraction(wavelength)
        assert isinstance(result, na.AbstractScalar)
        assert np.all(result > 0)

    @pytest.mark.parametrize("wavelength", _wavelength)
    def test_wavenumber(
        self,
        a: optika.chemicals.AbstractChemical,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.wavenumber(wavelength)
        assert isinstance(result, na.AbstractScalar)
        assert np.all(result >= 0)

    @pytest.mark.parametrize("wavelength", _wavelength)
    def test_absorption(
        self,
        a: optika.chemicals.AbstractChemical,
        wavelength: u.Quantity | na.AbstractScalar,
    ):
        result = a.absorption(wavelength)
        assert isinstance(result, na.AbstractScalar)
        assert np.all(result >= 0)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.chemicals.Chemical(
            formula=formula,
            is_amorphous=is_amorphous,
            table=table,
        )
        for formula in [
            "Si",
            "SiO2",
            na.ScalarArray(np.array(["Si", "SiO2"]), axes="layer"),
        ]
        for is_amorphous in [
            False,
            True,
        ]
        for table in [None, "palik"]
    ],
)
class TestChemical(
    AbstractTestAbstractChemical,
):
    pass



import pytest
import numpy as np
import astropy.units as u
import named_arrays as na
import optika
from .._tests import test_mixins
from optika.rays._tests import test_ray_vectors

_wavelength = na.linspace(100, 300, axis="wavelength", num=11) * u.AA


class AbstractTestAbstractRulings(
    test_mixins.AbstractTestPrintable,
    test_mixins.AbstractTestShaped,
):
    def test_diffraction_order(self, a: optika.rulings.AbstractRulings):
        assert np.issubdtype(na.get_dtype(a.diffraction_order), int)

    def test_spacing(self, a: optika.rulings.AbstractRulings):
        result = a.spacing
        types = (u.Quantity, na.AbstractScalar, optika.rulings.AbstractRulingSpacing)
        assert isinstance(result, types)

    def test_spacing_(self, a: optika.rulings.AbstractRulings):
        result = a.spacing_
        assert isinstance(result, optika.rulings.AbstractRulingSpacing)

    @pytest.mark.parametrize(
        argnames="rays",
        argvalues=test_ray_vectors.rays,
    )
    @pytest.mark.parametrize(
        argnames="normal",
        argvalues=[
            na.Cartesian3dVectorArray(0, 0, -1),
        ],
    )
    def test_efficiency(
        self,
        a: optika.rulings.AbstractRulings,
        rays: optika.rays.RayVectorArray,
        normal: na.AbstractCartesian3dVectorArray,
    ):
        result = a.efficiency(
            rays=rays,
            normal=normal,
        )

        assert np.all(result >= 0)
        assert np.all(result <= 1)


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.Rulings(
            spacing=1 * u.um,
            diffraction_order=1,
        ),
        optika.rulings.Rulings(
            spacing=1 * u.um,
            diffraction_order=na.ScalarArray(np.array([-1, 0, 1]), axes="m"),
        ),
    ],
)
class TestRulings(
    AbstractTestAbstractRulings,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.MeasuredRulings(
            spacing=1 * u.um,
            diffraction_order=1,
            efficiency_measured=na.FunctionArray(
                inputs=na.SpectralDirectionalVectorArray(
                    wavelength=_wavelength,
                    direction=na.Cartesian3dVectorArray(0, 0, 1),
                ),
                outputs=np.exp(-np.square(_wavelength / (10 * u.AA)) / 2),
            ),
        )
    ],
)
class TestMeasuredRulings(
    AbstractTestAbstractRulings,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.SinusoidalRulings(
            spacing=1 * u.um,
            depth=10 * u.nm,
            diffraction_order=na.ScalarArray(np.array([0, 1, 2]), axes="m"),
        ),
    ],
)
class TestSinusoidalRulings(
    AbstractTestAbstractRulings,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.SquareRulings(
            spacing=1 * u.um,
            depth=10 * u.nm,
            diffraction_order=na.ScalarArray(np.array([0, 1, 2]), axes="m"),
        ),
    ],
)
class TestSquareRulings(
    AbstractTestAbstractRulings,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.SawtoothRulings(
            spacing=1 * u.um,
            depth=10 * u.nm,
            diffraction_order=na.ScalarArray(np.array([0, 1, 2]), axes="m"),
        ),
    ],
)
class TestSawtoothRulings(
    AbstractTestAbstractRulings,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.TriangularRulings(
            spacing=1 * u.um,
            depth=10 * u.nm,
            diffraction_order=na.ScalarArray(np.array([0, 1, 2]), axes="m"),
        ),
    ],
)
class TestTriangularRulings(
    AbstractTestAbstractRulings,
):
    pass


@pytest.mark.parametrize(
    argnames="a",
    argvalues=[
        optika.rulings.RectangularRulings(
            spacing=1 * u.um,
            depth=10 * u.nm,
            ratio_duty=0.3,
            diffraction_order=na.ScalarArray(np.array([0, 1, 2]), axes="m"),
        ),
    ],
)
class TestRectangularRulings(
    AbstractTestAbstractRulings,
):
    pass
